!******************************************************
! August 2000
! Module to correct the astrometry for misalignment
! of the ra-dec axes of the XMM and usno catalogue.
! Will also determine the rotation between the 2 axes.
! Based on the original code in omatt_mod.90, but ammended
! so that it works.
!******************************************************
module astromcorrect_mod
    use param
    use dal
    use errorHandling
    use sla

    use sasdal
    use allocation_mod    
    use utility_mod  
    use sourcematch_mod
!    use median_mod
    use statistics_mod

public

    real(kind = double), public, parameter :: PI = 3.141592653589793_double  
    real(kind = double), parameter         :: radiansToDegrees = 180.0_double / PI
    real(kind = double), parameter         :: degreesToRadians = PI / 180.0_double
    real(kind = double), parameter         :: radiansToArcSecs = radiansToDegrees * 3600.0_double 

    real(kind=double), parameter           :: xCentre = 1024.5_double
    real(kind=double), parameter           :: yCentre = 1024.5_double
 
    !**************************************************************
    ! A structure for use in matching the sources to the catalogue
    !**************************************************************
    Type SHIFT
        real(kind=single) :: x
        real(kind=single) :: y
        real(kind=single) :: radius
    end Type SHIFT

    !*********************************************
   ! A structure for reading the catalogue data 
    !*********************************************
    type CatalogueData
        real(kind=double) :: RA
        real(kind=double) :: dec
        real(kind=single) :: error
        real(kind=single) :: magnitude
    end type CatalogueData
  
contains
!***************************************************************
! Astrometry Correction to star positions using a star catalogue
!***************************************************************
subroutine CorrectAstrometry(noStars, raSource, decSource, &
        posErr, sourceMagnitude, raPnt, decPnt, fileName, &
        status, xShift, sigXShift, yShift, sigYShift, &
        rotation, sigRotation, xScale, yScale, maxLikelihood, &
        noMatched)     
    integer, intent(in) :: noStars                   ! The number of stars
   
    real(kind=double), intent(inout), dimension(1 : noStars) :: raSource
    real(kind=double), intent(inout), dimension(1 : noStars) :: decSource 
    real(kind=single), intent(in), dimension(1 : noStars) ::    posErr 
    real(kind=single), intent(in), dimension(1 : noStars) :: sourceMagnitude 
    real(kind=double), intent(in) :: raPnt, decPnt
 
    character(len=*), intent(in) :: fileName        ! Used for getting various parameters
  
     
    logical, intent(out) :: status
  
    real(kind=single), intent(out)    :: xShift, yShift 
    real(kind=single), intent(out)    :: xScale, yScale 
    real(kind=single), intent(out)    :: rotation, sigRotation 
    real(kind=single), intent(out)    :: sigxShift, sigyShift 
    real(kind=double), intent(out)    :: maxLikelihood
    integer(kind=int32), intent(out) :: noMatched
    real(kind=double)    :: ep2     ! equinox
      
    type(DataSetT) :: dataSetH

    real(kind=double), pointer, dimension(: ,  : ) :: xye, xym      

    real(kind=double) :: raSourceMin, raSourceMax
    real(kind=double) :: decSourceMin, decSourceMax 
    real(kind=double) :: raCatMin, raCatMax
    real(kind=double) :: decCatMin, decCatMax 

    real(kind=double) :: raMin, decMin, raRange, decRange


    integer(kind=int32)  :: nCat                                ! number of catalogue stars

    real(kind=double), pointer :: xiSource(:), etaSource(:)
  
    real(kind=double), pointer :: xiCat(:), etaCat(:) 
    real(kind=double), pointer ::  separation(:)
    
   
    logical, allocatable ::  useSource(:) 
    character(len=30) :: obsId, expID, object
 
    integer :: i, j     
  
    integer, parameter          :: xCorrRange = 20 ! was 20 
    integer                     :: istat          ! status for memory allocation
    
   
    logical, allocatable  :: useCat(:)
    integer :: nCatInWindow = 0
    real(kind=single) :: tolerance = 3.0          ! Tolerance for catalog star search (arcsec) was 2.
    character(len=256)   :: catFileName
  
    
    integer(kind=int32), pointer :: sourceIndex(:)
    integer(kind=int32), pointer :: catIndex(:)
     
    real(kind=double), dimension(1 : 6) :: coeffs   
 
    logical :: okay
    real(kind=single), dimension(1 : 3) :: maxMagDifference = (/2.0, 1.0, 99999.0/) 
 
    integer, dimension(1 : 3)           :: nSigma = (/5, 5, 5/)

    real(kind=double) :: rmsResidual

    integer :: nStars
    integer :: count
    real(kind=single), dimension(1 : 3) ::  magLimit = (/15.0, 18.0, 25.0/)
    character(len=200) :: solutionString
    integer(kind=int32), dimension(1 : 3) :: noIntervals = (/1, 2, 4/)
    real(kind=single) :: medianPositionDifference
    real(kind=single) :: medianMagDifference
    integer :: iStart, iFinish, jStart, jFinish
    type(CatalogueData), pointer :: catData(:)

   
    catfileName = trim(stringParameter("catfile"))
    !********************************************************
    ! Get the catalogue data from the fits file (if present)
    !********************************************************
    open(unit=2, status = 'OLD', err = 1, file = catFileName)
    goto 2
 1  status = .false.
    call warning("Star catalogue not found- no astrometry correction done", catFileName) 
    return       ! File doesn't exist
 2  close(unit = 2)




    if (noStars .lt. 10) then 
        call warning("Too few sources for astrometry", "Need at least 10")
        status = .false.
        return
    end if


    !*********************************************
    ! Don't use stars with large position errors 
    ! (The error is in pixels)
    !*********************************************
    allocate(useSource(1 : noStars), stat = istat)
    call CheckAllocation(istat, "useSource")
    useSource = .TRUE.

    useSource = .false.
    nStars = 0

    do i = 1, noStars
        if (posErr(i) .le. tolerance) then
            useSource(i) = .true.
            nStars = nStars + 1
        end if 
    end do



    if (nStars .lt. 10) then 
        call warning("Too few sources for astrometry", "Need at least 10")
        status = .false.
        return
    end if



    !******************************
    ! Get the catalogue data
    !******************************

    dataSetH = DataSet(fileName, read)
  
    ep2 = real32Attribute(dataSetH, "EQUINOX")
 
    catData =>  GetCatalogueData(catFileName, ep2, nCat)
   

    if (nCat .eq. 0) then 
        call warning("noCat", "Catalogue has not been used to correct positions")
        status = .false.
        return
    end if
    if (maxval(catData%magnitude) .lt. 0.0) maxMagDifference = 99999.0
 
    xiSource => AllocateDouble(1, noStars, "xiSource")

    etaSource => AllocateDouble(1, noStars, "etaSource")
  
  
    !*********************************************************************
    ! Check for an overlap of the catalogue with the source positions
    !********************************************************************

    raSourceMin = minval(raSource) - (real(xcorrRange) / 3600.0) * degreesToRadians
    raMin = raSourceMin
    raSourceMax = maxval(raSource) + (real(xcorrRange) / 3600.0) * degreesToRadians
    
    decSourceMin = minval(decSource) - (real(xcorrRange) / 3600.0) * degreesToRadians
    decMin = decSourceMin
    decSourceMax = maxval(decSource) + (real(xcorrRange) / 3600.0) * degreesToRadians
   
    raRange = raSourceMax - raSourceMin
    decRange = decSourceMax - decSourceMin

   


    obsId = stringAttribute(dataSetH, "OBS_ID")
    expID = stringAttribute(dataSetH, "EXP_ID")
 

    object = stringAttribute(dataSetH, "OBJECT") 


    call release(dataSetH)

  

    raCatMin = catData(1)%RA
    raCatMax = catData(nCat)%RA

    decCatMin = 10000.0
    decCatMax = -10000.0
    do i = 1, nCat
        decCatMin = min(decCatMin, catData(i)%dec)
        decCatMax = max(decCatMax, catData(i)%dec)
    end do
        
    if (raCatMin .gt. raSourceMax .or. decCatMin .gt. decSourceMax .or. &
        raCatMax .lt. raSourceMin .or. decCatMax .lt. decSourceMin) then 
      
        call warning("Cannot do astrometry correction", "No overlap between sources and catalogue stars")
        status = .false.
        return
    end if
  
  


    allocate(useCat(1 : nCat), stat = istat)
    call CheckAllocation(istat, "useGs")

    useCat = .true.




    xiCat => AllocateDouble(1, nCat, "xiCat")
    etaCat => AllocateDouble(1, nCat, "etaCat")

 
    do i = 1, nCat

   
        !*********************************  
        ! Compute tangent plane coords
        ! xi_gs(i) & eta_gs(i)
        !*********************************
        call sla_ds2tp(catData(i)%RA, catData(i)%dec, raPnt, &
                  decPnt, xiCat(i), etaCat(i), j)

        !*********************************
        ! Check to see okay
        !********************************
        call SlaErrorCheck(j)
 
    end do


  
    !*****************************************************
    ! Create a new list of  the catalog stars that only 
    ! includes those that fall in the osw
    !*****************************************************

    nCatinWindow = 0
    do i = 1, nCat
        if (catData(i)%RA .gt. raSourceMin .and. catData(i)%RA .lt. raSourceMax .and. &
            catData(i)%dec .gt. decSourceMin .and. catData(i)%dec .lt. decSourceMax) then
            nCatInWindow = nCatinWindow + 1
            catData(nCatInWindow)%RA = catData(i)%RA
            catData(nCatInWindow)%dec = catData(i)%dec
            catData(nCatInWindow)%magnitude = catData(i)%magnitude
            xiCat(nCatInWindow) = xiCat(i)
            etaCat(nCatInWindow) = etaCat(i)

        end if
    end do

    nCat = nCatInWindow 
     
    if (nCat .lt. 10) then 
        call warning("Too few guide stars for astrometry correction", "Catalogue has not been used to correct positions")
        status = .false.
        return
    end if


    separation => AllocateDouble(1, nStars, "separation")
    separation = 1.0_double

    xye => Allocate2DDouble(1, 2, 1, nCat, "xye")
    xym => Allocate2DDouble(1, 2, 1, nCat, "xym")
 
  
    !******************************************  
    ! Compute the tangent plane coordinates
    ! of each source using the RAS and decs and
    ! approximate camera pointing direction.
    !******************************************
          
      
    do i = 1, noStars

        call sla_ds2tp(raSource(i), decSource(i), raPnt, &
                  decPnt, xiSource(i), etaSource(i), j)

        !*********************************
        ! Check to see okay
        !********************************
        call SlaErrorCheck(j)
 
    end do


    
    count = 0
    do i = 1, noStars
        if (sourceMagnitude(i) .gt. 0.0) count = count + 1
    end do
   

    if (real(count) / real(noStars) .lt. 0.5 .or. maxval(catData%magnitude) .lt. 0.0) then
    
        maxMagDifference = 99999.0
        magLimit = 25.0
    else
    
 
        magLimit(1) = 14.5
        count = 0    
        do while (count .lt. 20)
        count = 0
        magLimit(1) = magLimit(1) + 0.5
        if (magLimit(1) .gt. 25.0) exit
            do i = 1, noStars
                if (sourceMagnitude(i) .gt. 0.0 .and. sourceMagnitude(i) .le. magLimit(1)) count = count + 1
            end do
    
        end do
        if (magLimit(1) .ge. magLimit(2)) then    
            magLimit(2) = magLimit(3)
!            maxMagDifference = 99999.0
        end if
    end if
      

    !******************************************
    ! Find the x and y offsets that give the 
    ! best match between the sources and
    ! catalogue objects 
    !******************************************
    sourceIndex => AllocateInt32(1, noStars, "sourceIndex")
    
    catIndex => AllocateInt32(1, noStars, "catIndex")
 

    iStart = -xCorrRange
    iFinish = xCorrRange
    jStart = iStart
    jFinish = iFinish

    !**********************************************************
    ! Find the best offsets using smaller and smaller grid
    ! searches
    !*********************************************************
    do i = 1, 3

        sourceIndex = 0
        catIndex = 0

        call FindBestOffsets(noIntervals(i), iStart, iFinish, jStart, jFinish, &
                         noStars, raSource, decSource, posErr, &
                         sourceMagnitude, useSource, & 
                         nCat, catData%RA, catData%dec, catData%magnitude, magLimit(i), &
                         maxMagDifference(i), nSigma(i),  & 
                         noMatched, xShift, yShift, rmsResidual, maxLikelihood, &
                         medianPositionDifference, medianMagDifference, &
                         sourceIndex, catIndex)


        !**********************************************************
        ! Adjust the catalogue magnitudes for the magnitude offset
        !**********************************************************
        catData%magnitude = catData%magnitude - medianMagDifference
 

        if (noMatched .lt. 10) then 

       
            call warning("Too few sources matched for astrometry", "Need at least 10")
            status = .false.
            return
        end if 
        !***************************
        ! The new search box
        !***************************
       

        xShift = xShift * radiansToDegrees * 3600.0
        yShift = yShift * radiansToDegrees * 3600.0
         
        iStart = int(xShift)  - 3 
        iFinish = int(xShift) + 3
        jStart = int(yShift)  - 3
        jFinish = int(yShift) + 3

    end do




 
  

    forall (i = 1 : noMatched)
        
        xye(1, i) = xiCat(catIndex(i))    
        xye(2, i) = etaCat(catIndex(i))
        xym(1, i) = xiSource(sourceIndex(i))
        xym(2, i) = etaSource(sourceIndex(i))
    end forall       

      
    write(errstr, FMT='(a, f8.2, a, f8.2, a)') "Apparent image offset=", &
              xShift, ", ", yShift, " (arc secs)"
    
    write(errstr, FMT='(a, f8.2, a, f8.2, a)') "Apparent image offset=", &
              xShift, ", ", yShift, " (arc secs)"
  
    call message(AppMsg, Verbose)



    write(errstr, '(a, i4)') "Number of catalog stars in window: ", nCatInWindow
    call message(AppMsg, Verbose)
    write(errstr, '(a, i4)') "Number of objects in source list: ", noStars
    call message(AppMsg, Verbose)

    
    write(errstr,'(a, f5.1, a1)') "Fraction of sources matched: ", &
                    100.0 * real(noMatched) / real(noStars), "%"

    call message(AppMsg, Verbose)

    call message(AppMsg, Verbose)
    

    
    !***********************************************************
    ! Find the plate solution and associated errors (determined
    ! by Monte Carlo simulations)
    !***********************************************************
    do i = 1, 3
 
        write(errstr, FMT='(a, i1)') "Iteration no ", i
        call message(AppMsg, Verbose)             
 
        call Solution(noMatched, xye, xym, coeffs, status, &
                       xShift, sigxShift, yShift, sigYshift, &
                       xScale, yScale, rotation, sigRotation, &
                       solutionString)
        if (.not. status) exit
        !********************************************
        ! Find the goodness of  fit
        !********************************************
        call GoodnessOfFit(noMatched, xye, xym, coeffs, raPnt, decPnt, okay)
        if (.not. okay) exit
    end do
 

   


    if (status) then

     
        !***********************************
        ! Correct the source positions
        !***********************************


        call CorrectSourcePositions(nStars, raSource, decSource, &
                                    coeffs, raPnt, decPnt)

        !*******************************
        ! Show the fit
        !*******************************
        
        do  i = 1, noMatched
                      
            !**********************************************
            ! Correct these coordinates using the fitted
            ! parameters
            !*********************************************
 
            call sla_xy2xy(xym(1, i), xym(2, i), coeffs, xiSource(i), etaSource(i))
       
            xiCat(i)  = xye(1, i)
            etaCat(i) = xye(2, i)
            
        end do


!        call DisplayFit(noMatched, xiSource, etaSource, &
!                     xiCat, etaCat, obsId, expID, object, solutionString)

    else

        call warning("noFit", "No astrometric corrections applied")

    end if

 

    !**********************************
    ! Tidy up
    !**********************************


    deallocate(xiSource, stat = istat)
    call CheckDeallocation(istat, "xiSource")

    deallocate(etaSource, stat = istat)
    call CheckDeallocation(istat, "etaSource")

    deallocate(catData, stat = istat)
    call CheckDeallocation(istat, "catData")


    deallocate(separation, stat = istat)
    call CheckDeallocation(istat, "separation")


    deallocate(useSource, stat = istat)
    call CheckDeallocation(istat, "useSource")

    deallocate(xiCat, stat = istat)
    call CheckDeallocation(istat, "xiCat")

    deallocate(etaCat, stat = istat)
    call CheckDeallocation(istat, "xiCat")
  
    deallocate(useCat, stat = istat)
    call CheckDeallocation(istat, "useGs")

    deallocate(xye, stat = istat)
    call CheckDeallocation(istat, "xye")

    deallocate(xym, stat = istat)
    call CheckDeallocation(istat, "xym")
    deallocate(sourceIndex, stat = istat)
    call CheckDeallocation(istat, "sourceIndex")
    
    deallocate(catIndex, stat = istat)
    call CheckDeallocation(istat, "catIndex")


  
end subroutine CorrectAstrometry


!************************************************
! subroutine Solution
! 
! 
!***********************************************
subroutine Solution(npts, xye, xym, coeffs, fitOkay, &
                         xOffset, stdxOffset, yOffset, &
                         xScale, stdyOffset, yScale, &
                         orient, stdOrient, solutionString)
    integer, intent(in) :: npts
    real(kind=double), intent(in), dimension(1 : 2, 1 : npts) :: xye, xym
    real(kind=double), intent(out), dimension(1 : 6) :: coeffs
    real(kind=single), intent(out) :: xOffset, yOffset, xScale, yScale, orient
    real(kind=single), intent(out) :: stdxOffset, stdyOffset, stdOrient
    real(kind=single) :: stdxScale, stdyScale 
    real(kind=double)              :: xOffset1, yOffset1, xScale1, yScale1, orient1
    real(kind=double)              :: stdxOffset1, stdyOffset1, stdxScale1, stdyScale1, stdorient1
      
    real(kind=double), pointer :: xyeSim(:, :), xymSim(:, :)
    real(kind=double), pointer :: simCoeffs(:, :)
    real(kind=double), pointer :: data(:)
    real(kind=double), pointer :: error(:)
    real(kind=double) :: perp
    real(kind=double) :: number
    integer i, j, k, index, nData, nSim
    integer :: istat
    logical, intent(out) :: fitOkay
    character(len=*), intent(out) :: solutionString


  
    !*************************************************
    ! Set the number of simulations to 10 x data size,
    ! or a maximum of 1000
    !*************************************************
    nSim = min(1000, 10 * npts)

    xyeSim => Allocate2DDouble(1, 2, 1, npts, "xyeSim")
    xymSim => Allocate2DDouble(1, 2, 1, npts, "xymSim")
    simCoeffs => Allocate2DDouble(1, 5, 1, nSim, "simCoeffs")
    nData = 0

    do i = 1, nSim
        do j = 1, 2
            do k = 1, npts
                call random_number(number)
                index = int(number * real(npts - 1) + 0.0001) + 1
                xyeSim(j, k) = xye(j, index)
                xymSim(j, k) = xym(j, index)
            end do
        end do
        !********************************************************
        ! Find the plate constants for this simulation
        !*******************************************************     
        call PlateSolution(npts, xyeSim, xymSim, coeffs, fitOkay)
        if (fitOkay) then
            nData = nData + 1

            !*******************************************************
            ! decompose the fit to offsets, scales and orientations
            !*******************************************************
            call sla_dcmpf(coeffs, simCoeffs(1, nData),      & 
                   simCoeffs(2, nData), simCoeffs(3, nData), & 
                   simCoeffs(4, nData), perp, &
                   simCoeffs(5, nData))
               
        end if
    end do                       

    if (.not. fitOkay) return
 
    !********************************************************
    ! Find the mean and standard deviation of the simulated
    ! fitted coefficients
    !********************************************************
    error => AllocateDouble(1, nData, "error")
    error = 0.0_double 
    data => AllocateDouble(1, nData, "data")
     
    forall(i = 1 : nData) data(i) = simCoeffs(1, i) 
    call ComputeAverage(data, error, nData, xOffset1, stdxOffset1)

   
    forall(i = 1 : nData) data(i) = simCoeffs(2, i) 
    call ComputeAverage(data, error, nData, yOffset1, stdyOffset1)

    forall(i = 1 : nData) data(i) = simCoeffs(3, i) 
    call ComputeAverage(data, error, nData, xScale1, stdxScale1)

    forall(i = 1 : nData) data(i) = simCoeffs(4, i) 
    call ComputeAverage(data, error, nData, yScale1, stdyScale1)

    forall(i = 1 : nData) data(i) = simCoeffs(5, i) 
    call ComputeAverage(data, error, nData, orient1, stdOrient1)


    xOffset = real(xOffset1)
    stdxOffset = real(stdxOffset1)

      

    yOffset = real(yOffset1)
    stdyOffset = real(stdyOffset1)


    xScale = real(xScale1)
    stdxScale = real(stdxScale1)

    yScale = real(yScale1)
    stdyScale = real(stdyScale1)

    orient = real(orient1)
    stdOrient = real(stdOrient1)
 
       
    write(errstr, '(a, i4, a)') 'Plate solution (no=', npts, '): '
   
    solutionString = trim(errstr) 
    call message(AppMsg, Verbose)
    xOffset = xOffset * radiansToArcSecs
    stdxOffset = stdxOffset *  radiansToArcSecs


    write(errstr, '(a, f6.2, a, f6.2, a)') 'dx=', xOffset, "+/-", stdxOffset, " (arcsecs)"
    solutionString = trim(solutionString) // " " // trim(errstr)   
    call message(AppMsg, Verbose)

    yOffset = yOffset * radiansToArcSecs
    stdyOffset = stdyOffset * radiansToArcSecs

    write(errstr, '(a, f6.2, a, f6.2, a)') 'dy=', yOffset, "+/-", stdyOffset, " (arcsecs)"
    solutionString = trim(solutionString) // " " // trim(errstr)   
      
    call message(AppMsg, Verbose)
    write(errstr, '(a, f7.4, a, f7.4 )') 'xscale=', xScale, "+/-", stdxScale
  
    call message(AppMsg, Verbose)
    write(errstr, '(a, f7.4, a, f7.4 )') 'yscale=', yScale, "+/-", stdyScale
   
    call message(AppMsg, Verbose)
    orient = orient * radiansToDegrees
    stdOrient = stdOrient * radiansToDegrees
     
    write(errstr, '(a, f6.3, a, f6.3, a)') 'rotation=', orient, "+/-", stdOrient, " (degrees)"
    solutionString = trim(solutionString) // " " // trim(errstr)   


      
    call message(AppMsg, Verbose)


    deallocate(xyeSim, stat = istat)
    call CheckDeallocation(istat, "xye")

    deallocate(xymSim, stat = istat)
    call CheckDeallocation(istat, "xym")

    deallocate(simCoeffs, stat = istat)
    call CheckDeallocation(istat, "simCoeffs")

    deallocate(data, stat = istat)
    call CheckDeallocation(istat, "data")
   
    deallocate(error, stat = istat)
    call CheckDeallocation(istat, "error")
   

end subroutine Solution 
!************************************************
! subroutine PlateSolution
!***********************************************
subroutine PlateSolution(npts, xye, xym, coeffs, fitOkay)
    integer, intent(in) :: npts
    real(kind=double), intent(in), dimension(1 : 2, 1 : npts) :: xye, xym
    real(kind=double), intent(out), dimension(1 : 6) :: coeffs
 
    integer j
    logical, intent(out) :: fitOkay
    

    !*******************************************************************
    ! call linear fit, assume that x,y are perpendicular (i.e. no shear)
    !******************************************************************* 
    call sla_fitxy(4, npts, xye, xym, coeffs, j)

    !********************************
    ! check that the fit was ok
    !********************************
    select case (j)
        case(-2)
            call warning("insuffData",&
                  "insufficient guide star data for astrometric fit")

            fitOkay = .false.
    
            return
        case(-3)
            call warning("singularSol","Singular solution to astrometric fit")
     
            fitOkay = .false.
            return
        case default

            fitOkay = .true.
    end select
 
     
end subroutine PlateSolution 

!******************************************
! Subroutine CorrectSourcePositions
! Corrects the standard coordinates
!******************************************
subroutine CorrectSourcePositions(nStars, raSource, decSource, &
                                  coeffs, &
                                  raPnt, decPnt)
    integer, intent(in) :: nStars
    real(kind=double), intent(inout), dimension(1 : nStars) :: raSource
    real(kind=double), intent(inout), dimension(1 : nStars) :: decSource
 
    real(kind=double), dimension(1 : 6) :: coeffs
    real(kind=double), intent(in) :: raPnt, decPnt
    integer i, j
    real(kind=double) :: x1, x2, y1, y2
   
    write(errstr, '(a)') "Correcting source positions"
    call message(AppMsg, Verbose)

    !***************************************************
    ! correct the source positions using the fitted 
    ! parameters
    !***************************************************
    do i = 1, nStars

        !*********************************************
        ! Calaculate the standard coordinates for this 
        ! object
        !*********************************************
        call sla_ds2tp(raSource(i), decSource(i), raPnt, &
                  decPnt, x1, y1, j)

        !*********************************
        ! Check to see okay
        !********************************
        call SlaErrorCheck(j)
 
        !**********************************************
        ! Correct these coordinates using the fitted
        ! parameters
        !********************************************* 
        call sla_xy2xy(x1, y1, coeffs, x2, y2)
       
        !***********************************************
        ! Convert the corrected standard coordinates
        ! to RA and DEC
        !***********************************************
        call sla_dtp2s(x2, y2, raPnt, decPnt, &
                       raSource(i), decSource(i))       
        
    end do
end subroutine CorrectSourcePositions
!**********************************
! subroutine SlaErrorCheck
!**********************************
subroutine SlaErrorCheck(j)
    integer, intent(in) :: j

    select case (j)
        case (1) 
            call error("toofar", "Guide star too far from axis")
        case (2)
            call error("antistar", "Anti-star on tangent plane")
        case (3)
            call error("toofar", "Anti-star too far from axis")
    end select
end subroutine SlaErrorCheck
!************************************************************
! Subroutine SortOffsets
! Sorts  a i-d array of objects of type SHIFT into increasing
! radius order
!************************************************************
subroutine SortOffsets(array, size)
    integer(kind = int32), intent(in) :: size
    type(SHIFT), intent(inout), dimension(1 : size) :: array
    integer(kind = int32) :: i, ir, j, l
    type(SHIFT) :: rra
  
    if (size .lt. 2) return

    l = size / 2 + 1
    ir = size   

    do 
        if (l .gt. 1) then
            l = l - 1
            rra = array(l)    
        else
            rra = array(ir)
            array(ir) = array(1)
            ir = ir - 1
            if (ir .le. 1) then
                array(1) = rra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j .le. ir)
       
            if (j .lt. ir) then
                if(array(j)%radius .lt. array(j + 1)%radius) j = j + 1
            end if
            if (rra%radius .lt. array(j)%radius) then
                array(i) = array(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
        array(i) = rra    
    end do
end subroutine SortOffsets

!**********************************************************
! Subroutine DisplayFit
!
!*********************************************************
subroutine DisplayFit(size, raSource, decSource, &
                     raCat, decCat,  &
                     obsId, expID, object, solutionString)
    integer, intent(in) :: size
    
    real(kind=double), intent(in), dimension(1 : size) :: raSource
    real(kind=double), intent(in), dimension(1 : size) :: decSource
    real(kind=double), intent(in), dimension(1 : size) :: raCat
    real(kind=double), intent(in), dimension(1 : size) :: decCat
 
    character(len=*), intent(in) :: obsId, expID, object, solutionString
    real(kind=double), pointer :: separation(:)
 
    real(kind=double), pointer :: error(:)
    
    real(kind=single), pointer :: residual(:)
    real(kind=single), pointer :: raOffset(:)
    real(kind=single), pointer :: decOffset(:)
    real(kind=single), pointer :: error1(:)



    real(kind=double) :: mean, standardDeviation
    integer :: allocErr
    character(len=100) :: xlabel, title
    integer :: noBIns
    character(len=200) :: comment, comment1
    character(len=50) :: fileName
    real(kind=single) :: binWidth
    integer :: i    
    real(kind=single) xmin, xmax, ymin, ymax
    !******************************************
    ! Plot out the residuals 
    !******************************************
 
    fileName = expID(1 : 10) // "_" // expID(11 : 13) // ".ps/VC"
 
    call pgbegin(0, trim(fileName), 1, 1)
    
    call pgsubp(1, 3)
    call pgpage
  
    separation => AllocateDouble(1, size, "separation")
    residual => AllocateSingle(1, size, "residual")
    error => AllocateDouble(1, size, "error")

    call ComputeResiduals(size, raSource, raCat, residual, mean, standardDeviation)
    

    write(comment, '(a, f8.4, a, f8.4, a)') "Mean=", mean, &
                  " (arc secs) Standard deviation=", standardDeviation, " (arc secs)"

    noBins = size / 10
    noBins = min(noBins, 20)    
    xLabel = "RA CORRECTED - RA CATALOGUE STAR (ARC SECS)"
   
    title = "ASTROMETRY CORRECTION (OMATT)"
    binWidth = 0.5
    call Histogram(size, residual, binWidth, &
                 xLabel, "", 0.1, 0.9, 0.1, 0.8)

!    call DrawNormalDistribution(size, dble(binWidth), mean, standardDeviation)
  
    call pgsch(1.5)
    call pgmtxt("t", 3.0, 0.5, 0.5, trim(title))
  

    call pgsch(1.2)
    comment1 = "OBS-ID=" // trim(obsId) // " OBJECT=" // trim(object) //  " EXPOSURE " // expID(11 : 13) 
    call pgmtxt("t", 2.0, 0.5, 0.5, trim(comment1))
       
  

    call pgsch(1.2)
   
    call pgmtxt("t", 1.0, 0.5, 0.5, trim(solutionString))

    write(comment, '(a, f8.4, a, f8.4, a)') "Mean=", mean, &
                  " (arc secs) Standard deviation=", standardDeviation, " (arc secs)"
    call pgmtxt("t", -2.0, 0.5, 0.5, trim(comment))
  

    call pgpanl(1, 2)
    xLabel = "DEC CORRECTED - DEC CATALOGUE STAR (ARC SECS)"
   
    call ComputeResiduals(size, decSource, decCat, residual, mean, standardDeviation)
    
    binWidth = 0.5
    call Histogram(size, residual, binWidth, &
                 xLabel, "", 0.1, 0.9, 0.1, 0.8)

!    call DrawNormalDistribution(size, dble(binWidth), mean, standardDeviation)
      
    call pgsch(1.2)
   
    write(comment, '(a, f8.4, a, f8.4, a)') "Mean=", mean, &
                  " (arc secs) Standard deviation=", standardDeviation, " (arc secs)"
    call pgmtxt("t", -2.0, 0.5, 0.5, trim(comment))
  

    call pgpanl(1, 3)

    !*************************************
    ! A histogram of the separations
    !*************************************
    do i = 1, size
        separation(i) = ArcLength(raSource(i), decSource(i), & 
                        raCat(i), decCat(i))
        residual(i) = real(separation(i))
    end do 

        
    error = 1.0_double
    call ComputeAverage(separation, error, size, mean, standardDeviation)

    xLabel = "CORRECTED POSITION DEVIATIONS (ARC SECS)"
    binWidth = 0.5
    call Histogram(size, residual, binWidth, &
                 xLabel, "", 0.1, 0.9, 0.1, 0.8)

         
 
    call pgsch(1.2)
   

    write(comment, '(a, f8.4, a, f8.4, a)') "Mean=", mean, &
                  " (arc secs) Standard deviation=", standardDeviation, " (arc secs)"
    call pgmtxt("t", -2.0, 0.5, 0.5, trim(comment))
    call pgpage 

    raOffset => AllocateSingle(1, size, "raOffset")
    decOffset => AllocateSingle(1, size, "raOffset")
    error1 => AllocateSingle(1, size, "error1")

    forall (i = 1 : size)
        raOffset(i) = (raSource(i) - raCat(i)) *  radiansToDegrees * 3600.0_double
        decOffset(i) = (decSource(i) - decCat(i)) * radiansToDegrees * 3600.0_double
        error1(i) = 0.0
    end forall
    call pgsubp(1, 1)
    call pgpage
  
    call DrawGraph(size, raOffset, error1, decOffset, error1, &
         title, "RA SOURCE - RA CAT (arc secs)", "DEC SOURCE - DEC CAT (arc secs)", 0.1, 0.9, 0.1, 0.9)

    call pgsch(1.0)
 
!*******************************************
! For a test store offsets
!*******************************************
!    open(unit=1, status="old", err=1, position="append", file="omatt.dat")
!   goto 2
!1  open(unit=1, status="new", file="omatt.dat")
!   write(1, *) "Omatt astrometry correction offsets - RA & dec (arc secs)"  
!2 continue
!    do i = 1, size
!        write(1, 100) raOffset(i), decOffset(i)
!    end do
!100 format(2f8.2)

!    close(unit=1)

    call pgmtxt("t", 1.0, 0.5, 0.5, trim(comment1))
    call pgqwin(xmin, xmax, ymin, ymax)
    call pgmove(0.0, ymin)
    call pgdraw(0.0, ymax)
    call pgmove(xmin, 0.0)
    call pgdraw(xmax, 0.0)      
  

    call pgclos    
    deallocate(separation, stat = allocerr)
    deallocate(residual, stat = allocerr)
    deallocate(error, stat = allocerr)
    deallocate(raOffset, stat = allocerr)
    deallocate(decOffset, stat = allocerr)
    deallocate(error1, stat = allocerr)


          
end subroutine DisplayFit
!**************************************
! Subroutine ComputeResiduals
!*************************************
subroutine ComputeResiduals(size, obsData, fitData, residual, &
                           mean, standardDeviation)
    integer, intent(in) :: size
    real(kind=double), dimension(1 : size) :: obsData, fitData
    real(kind=single), intent(out), dimension(1 : size) :: residual
    real(kind=double), intent(out) :: mean, standardDeviation
    real(kind=double) :: diff, sum1
    integer :: i

    mean = 0.0_double
    
    do i = 1, size
         diff = obsData(i) - fitData(i)
         residual(i) = diff * radiansToDegrees * 3600.0_double
         mean = mean + diff   
    end do
    mean = (mean / dble(size))  * radiansToDegrees * 3600.0_double

    sum1 = 0.0_double
    

    do i = 1, size
        diff = mean - residual(i)
        sum1 = sum1 + diff * diff 
    end do
  
    standardDeviation = dsqrt(sum1 / dble(size - 1))  

end subroutine ComputeResiduals
!*****************************************************
! Subroutine Histogram
!
!*****************************************************
subroutine Histogram(ndata, data, binWidth, &
                 xLabel, title, x1, x2, y1, y2)
    integer, intent(in) :: ndata
    real(kind=single), intent(in), dimension(ndata) :: data
    real(kind=single), intent(inout) :: binWidth
    character(len=*), intent(in) :: title, xLabel
    real, intent(in) :: x1, x2, y1, y2
    real xmin, xmax, zmax, charSize, yMax    
    integer i, noBins
    integer(kind=int32), pointer:: frequency(:) 
  
    integer :: k = 1
    character(len=100) :: yLabel 
    xmin = minval(data)
    xmax = maxval(data)

    if (abs(xmin - xmax) .lt. 1e-6) then
        call warning("Unable to draw histogram", "xmin=xmax")
        return
    end if

    charSize = (x2 - x1) * 1.5

    xmin = 1e6
    xmax = -1e6
    do i = 1, ndata
        xmin = min(xmin, data(i))
        xmax = max(xmax, data(i))
    end do

    if (xmin .ge. 0.0 .and. xmax .gt. 0.0) then
        xmax = real(ceiling(xmax))
        xmin = real(floor(xmin))
    else
        zmax = max(abs(xmin), xmax)
        xmax = real(ceiling(xmax))
        xmin = -xmax
    end if
    
    do 
        noBins = int((xMax - xMin) / binWidth + 0.00001)
        if (noBins > 4) exit
        binWidth = binWidth / 2.0
    end do


    !******************************
    ! Find the maximum bin size
    !******************************

    frequency => AllocateInt32(1, noBins, "frequency")
    frequency = 0
    do i = 1, ndata
        k = int((data(i) - xmin) / binWidth + 0.00000001) + 1       
        k = max(1, k)
        k = min(noBins, k)
        frequency(k) = frequency(k) + 1
       
    end do

    yMax = int(real(maxval(frequency)) * 1.50) 

    deallocate(frequency)  
     
    call pgsvp(x1, x2, y1, y2)

    call pgwindow(xmin, xmax, 0.0, yMax)
   
     
    call pgsch(1.0) 
    call pgsfs(2)    
    call pgslw(5)    
    call pghist(ndata, data, xmin, xmax, noBins, 3)

    call pgwindow(xmin, xmax, 0.0, (yMax / real(ndata)) * 100.0)
  

    
    
    call pgbox("bctsi", binWidth, 5, "bctsi", 0.0, 0)

    call pgsch(1.0)  
    call pgbox("n", binWidth, 5, "n", 0.0, 0) 
  

    call pgsch(1.2)
    write(yLabel, "(a, i5, a)") "FREQUENCY (%) (TOTAL=", ndata, ")"
       
    call pglabel(xLabel, yLabel, title)
    

          
end subroutine Histogram
!*******************************************
! Subroutine SortCatalogue
!*******************************************
subroutine SortCatalogue(size, ra, dec, mag)
    integer, intent(in) :: size
    real(kind=double), intent(inout), dimension(1 : size) :: ra
    real(kind=double), intent(inout), dimension(1 : size) :: dec
    real(kind=single), intent(inout), dimension(1 : size) :: mag
    real(kind=double), pointer :: temp(:)
    integer(kind=int32), pointer :: index(:)
    
    integer :: i, status

    temp => AllocateDouble(1, size,  "temp/SortCatalogue")

    index => AllocateInt32(1, size,  "index/SortCatalogue")

    call HeapSort(ra, index, size)
    
    temp = ra
    forall(i = 1 : size) ra(i) = temp(index(i))

    temp = dec
    forall(i = 1 : size) dec(i) = temp(index(i))

    temp = mag
    forall(i = 1 : size) mag(i) = real(temp(index(i)))

    deAllocate(temp, stat = status)

    call CheckdeAllocation(status, "size/SortCatalogue")
    deAllocate(index, stat = status)
    call CheckdeAllocation(status, "index/SortCatalogue")





end subroutine SortCatalogue

!*****************************************
! Subroutine SetUpOffsets
! Sets up an array of ra-dec offsets to be
! used when matching sources with catalogue
! stars
!*****************************************
subroutine SetUpOffsets(noIntervals, iStart, iFinish, jStart, &
                         jFinish, noOffsets, offset)
    integer, intent(in) :: noIntervals
    integer, intent(in) :: iStart, iFinish, jStart, jFinish
    integer, intent(out) :: noOffsets
    type(SHIFT), intent(inout), dimension(:) :: offset 
    integer i, j
    integer :: iMax, jMax
    real :: x, y
    real :: radius
    real(kind=single) :: constant
        
    constant = degreesToRadians / 3600.0
   
    noOffsets = 0
    iMax = (abs(iStart) + abs(iFinish)) * noIntervals + 1
    jMax = (abs(jStart) + abs(jFinish)) * noIntervals + 1
  
    do i = 1, iMax
        x = real(iStart) + real(i - 1) / real(noIntervals)

        do j = 1, jMax
            y = real(jStart) + real(j - 1) / real(noIntervals)
    
            radius  = sqrt(x * x + y * y) 
            noOffsets = noOffsets + 1
            offSet(noOffsets)%x = x * constant
            offSet(noOffsets)%y = y * constant
            offSet(noOffsets)%radius = radius
            
        end do
    end do
 

    !*************************************************
    ! Sort these offsets into increasing radius order
    !************************************************
    call SortOffsets(offset, noOffsets)
  
 
end subroutine SetUpOffSets
!*************************************
! Function GetCatalogue data
! Reads in the RA, dec and magnitudes
! from the fits catalogue file
!*************************************
function GetCatalogueData(fileName, obsEpoch, nCat) result(catData)
    character(len=*), intent(in) :: fileName
    real(kind=double), intent(in) :: obsEpoch
    integer(kind=int32), intent(out) :: nCat   
    type(CatalogueData), dimension(:), pointer  :: catData
  
 
    type(DataSetT) :: catDataSetH
    type(TableT) :: inTableH
    integer :: i, n, n1, n2
    real(kind=double)    :: ep1
   

    ep1 = 2000.0_double
    
  

    n1 = 0 
    n2 = 0
    catDataSetH = dataSet(fileName, read)      ! 
    
    inTableH = table(catDataSetH, "RAWRES")

    nCat = numberOfRows(inTableH)


    allocate(catData(1 : nCat + 20))

    call ReadReal64Column(inTableH, "CAT_RA", nCat, 0, catData%RA)
    call ReadReal64Column(inTableH, "CAT_DEC", nCat, 0, catData%dec)

    if (HasColumn(inTableH, "BMAG")) then
     
        call ReadReal32Column(inTableH, "BMAG", nCat, 0, catData%magnitude)
        
    else if (HasColumn(inTableH, "RMAG")) then
       
        call ReadReal32Column(inTableH, "RMAG", nCat, 0, catData%magnitude)
    else 
        catData%magnitude = -999.0
    end if
    call release(inTableH)
    call release(catDataSetH)
  
    catData%error = 1.0  ! Set the position error to 1 arcsec    
 
    !******************************************
    ! Precess the catalogue coordinates to the 
    ! observation date
    !*********************************************
    catData%RA  = catData%RA * degreesToRadians
    catData%dec = catData%dec * degreesToRadians



    do i = 1, nCat

        !***************************************
        ! precess the gs coords to observation epoch
        !***************************************
        call sla_preces("FK5", ep1, obsEpoch, catData(i)%RA, catData(i)%dec)

    end do


    !*********************************************
    ! Sort the catalogue into RA increasing order
    !********************************************
    call SortCatalogue(nCat, catData%RA, catData%dec, catData%magnitude)


    do i = nCat + 10,  11, - 1
        catData(i) = catData(i - 10)
    end do



    do i = 1, 10
        catData(i)%RA = -1.0
        catData(10 + i + nCat)%RA = -1.0
    end do


    !*********************************************************
    ! Need to check to see if there are RAS near zero degrees
    !*********************************************************
    n1 = 0
    if (catData(11)%RA .lt. 0.006) then
        n1 = min(10, nCat)

        do i = 1, n1
            catData(i) = catData(nCat - n1 + i)
            catData(i)%RA =  catData(i)%RA - 360.0
     
        end do
       

    end if  
    

    !*******************************************************
    ! Need to check to see if there are RAS near 360 degrees
    !*******************************************************
    if (catData(nCat + 10)%RA .gt. 359.994) then
        n2 = min(10, nCat)
         do i = 1, n2
             catData(nCat + i) = catData(i + n1)
             catData(nCat + i)%RA =  catData(nCat + i)%RA + 360.0
         end do
    
    end if  
    
    n = 0
    do i = 1, nCat + 20
        if (catData(i)%RA .ge. 0.0) then
            n = n + 1
            catData(n) = catData(i)
        end if
    end do   
    nCat = n


end function GetCatalogueData
!******************************************
! Subroutine FindBestOffsets 
!******************************************
subroutine FindBestOffsets(noIntervals, iStart, iFinish, &
                           jStart, jFinish, &
                           noStars, raSource, decSource, posErr, &
                           sourceMagnitude, useSource, & 
                           nCat, raCat, decCat, magCat, magLimit, &
                           maxMagDifference, nSigma, & 
                           maxNoFound, xShift, yShift, rmsResidual, &
                           maxLikelihood, medianPositionDifference, &
                           medianMagDifference, sourceIndex, catIndex)

    integer, intent(in) :: noIntervals
    integer, intent(in) :: iStart
    integer, intent(in) :: iFinish
    integer, intent(in) :: jStart
    integer, intent(in) :: jFinish
    integer, intent(in) :: noStars
    real(kind=double), intent(in), dimension(1 : noStars) :: raSource
    real(kind=double), intent(in), dimension(1 : noStars) :: decSource 
    real(kind=single), intent(in), dimension(1 : noStars)    :: posErr 
    real(kind=single), intent(in), dimension(1 : noStars)    :: sourceMagnitude 
    logical, intent(in), dimension(1 : noStars)              ::  useSource 
    integer(kind=int32), intent(in) :: nCat 
    real(kind=double), intent(in), dimension(1 : nCat)  :: raCat     ! RAs of the catalogue stars
    real(kind=double), intent(in), dimension(1 : nCat)  :: decCat    ! decs of the catalogue stars
    real(kind=single), intent(in), dimension(1 : nCat) ::  magCat    ! mags of the catalogue stars 
    real(kind=single), intent(in) ::  magLimit
    real(kind=single), intent(in) ::  maxMagDifference
    integer, intent(in) :: nSigma
    integer, intent(out) :: maxNoFound
    real(kind=single), intent(out) :: xShift, yShift 
    real(kind=double), intent(out) :: rmsResidual
    real(kind=double), intent(out) :: maxLikelihood

    real(kind=single), intent(out) :: medianPositionDifference
    real(kind=single), intent(out) :: medianMagDifference
    integer, intent(out), dimension(1 : noStars) :: sourceIndex
    integer, intent(out), dimension(1 : noStars) :: catIndex

    !*********************************
    ! Local variables
    !********************************* 
    real(kind=double), pointer :: magnitudeOffset(:)
    real(kind=double), pointer :: separation(:)
    real(kind=single) :: fractionFound
    real(kind=single) :: mindist                  ! distances to GS
    real(kind=single) :: tolerance          ! Tolerance for catalog star search (arcsec) was 2.
    real(kind=single) :: spacing
    real(kind=single) :: xiShift
    real(kind=single) :: etaShift   
    real(kind=double) :: sum
    real(kind=double) :: likeLi
   
    real(kind=double) :: weight
    real(kind=double) :: sumWeight
    real(kind=double) :: likelihood
    real(kind=double) :: sigmaSquared    
    integer :: noMatched
    integer :: i
    integer :: k, l
    integer :: noOffsets
    integer :: iStat
    Type(SHIFT), dimension(:), allocatable  :: offSet 


    
    spacing = 1.0 / real(noIntervals)
    !******************************************************
    ! Set up the ra and dec offsets to be used in matching
    ! the sources with the catalogue stars
    !******************************************************
    noOffsets = ((abs(iStart) + abs(iFinish) + 1) * noIntervals) * &
               ((abs(jStart) + abs(jFinish) + 1) * noIntervals)  
    

    allocate(offSet(1 : noOffsets), stat = istat)
    call CheckAllocation(istat, "offSet")


    call SetUpOffsets(noIntervals, iStart, iFinish, jStart, jFinish, noOffsets, offset)
    magnitudeOffset => AllocateDouble(1, noStars, "magnitudeOffset")
    separation => AllocateDouble(1, noStars, "position")

    maxNoFound = 0
    rmsResidual = 999.9
    fractionFound = 0.0
    maxLikelihood = 0.0_double   

    !********************************************************
    ! Loop through all the offsets to find the best match
    ! between the sources and catalogue objects
    !********************************************************

    do i = 1, noOffsets 
 
        xiShift = offSet(i)%x
        etaShift = offset(i)%y


        noMatched = 0
        sum = 0.0_double
        sumWeight = 0.0_double        
        likelihood = 0.0_double
        do k = 1, noStars       
            if (.not. useSource(k) .or. sourceMagnitude(k) .gt. magLimit) cycle

            !**************************************************
            ! Find the nearest catalogue object to the source,
            ! checking that the magnitude difference is within
            ! the allowed range
            !**************************************************

            tolerance = real(nSigma) * posErr(k)
            if (SourceMatch(raSource(k) + dble(xiShift), decSource(k) + dble(etaShift), &
                          sourceMagnitude(k), &
                          nCat, raCat, decCat, magCat, tolerance, &
                          maxMagDifference, minDist, l, likeLi)) then
                
                noMatched = noMatched + 1            ! The number of sources matched       
                
                sigmaSquared = posErr(k) * posErr(k)
  
                weight = 1.0_double / sigmaSquared
                sum = sum + weight * minDist * minDist
                sumWeight = sumWeight + weight
                likelihood = likelihood + likeLi 

            end if

        end do
  
!        if (noMatched .gt. maxNoFound) then
        if (likelihood .gt. maxLikelihood) then

            sum = sqrt(sum / sumWeight)
!            if (sum .lt. rmsResidual) then
                rmsResidual = sum
                maxNoFound = noMatched
                xShift = xiShift
                yShift = etaShift          
                maxLikelihood = likelihood
                if (noMatched .gt. 3) then 
                    medianPositionDifference = MedianValue(separation, noMatched)
                    medianMagDifference = MedianValue(magnitudeOffset, noMatched)
                else
                    medianPositionDifference = 0.0
                    medianMagDifference = 0.0
                end if
              
                fractionFound = real(noMatched) / real(noStars)
                           
            end if        
!        end if
    end do
    noMatched = 0

    do k = 1, noStars       
        if (.not. useSource(k) .or. sourceMagnitude(k) .gt. magLimit) cycle

        !**************************************************
        ! Find the nearest catalogue object to the source,
        ! checking that the magnitude difference is within
        ! the allowed range
        !**************************************************

        tolerance = real(nSigma) * posErr(k)
        if (SourceMatch(raSource(k) + dble(xShift), decSource(k) + dble(yShift), &
             sourceMagnitude(k), nCat, raCat, decCat, magCat, tolerance, &
             maxMagDifference, minDist, l, likeLi)) then
                
            noMatched = noMatched + 1            ! The number of sources matched       
            magnitudeOffset(noMatched) = magCat(l) - sourceMagnitude(k)
            separation(noMatched) = minDist
            sourceIndex(noMatched) = k
            catIndex(noMatched) = l 
            

        end if
    end do
  
    deallocate(magnitudeOffset)
    deallocate(separation)
    deallocate(offset)
    

    
end subroutine FindBestOffsets
!************************************************
! subroutine GoodnessOfFit
! xye(1, i) - xi catalogue
! xye(2, i) - eta catalogue
! xym(1, i) - xi source
! xym(2, i) - eta source
! coeffs(1-6) - the fitted coefficients
!***********************************************
subroutine GoodnessOfFit(npts, xye, xym, coeffs, raPnt, decPnt, okay)
    integer, intent(inout) :: npts
    real(kind=double), intent(inout), dimension(1 : 2, 1 : npts) :: xye, xym
    real(kind=double), intent(in), dimension(1 : 6) :: coeffs
    real(kind=double), intent(in) :: raPnt, decPnt
    logical, intent(out) :: okay
    real(kind=single), pointer :: raSource(:)
    real(kind=single), pointer :: decSource(:)
    real(kind=single), pointer :: raCat(:)
    real(kind=single), pointer :: decCat(:)
    real(kind=double), pointer :: distance(:)
    real(kind=double), pointer :: error(:)

    real(kind=double) :: x, y, x1, y1
    real(kind=double) :: mean, standardDeviation
    real(kind=double) :: rmsResidual

    integer :: i, k, count
    integer :: nSigma

    
    raSource => AllocateSingle(1, npts,  "raSource")
    decSource => AllocateSingle(1, npts,  "decSource")

    raCat => AllocateSingle(1, npts,  "raCat")
    decCat => AllocateSingle(1, npts,  "decCat")
    distance => AllocateDouble(1, npts,  "distance")
    error => AllocateDouble(1, npts,  "error")



    rmsResidual = 0.0_double
    !***************************************************
    ! correct the source positions using the fitted 
    ! parameters
    !***************************************************
    do i = 1, npts


        !**********************************************
        ! Correct the source coordinates
        ! 
        !********************************************* 
        call sla_xy2xy(xym(1, i), xym(2, i), coeffs, x, y)
       
        !***********************************************
        ! Convert the corrected standard coordinates
        ! to RA and DEC
        !***********************************************
        call sla_dtp2s(x, y, raPnt, decPnt, x1, y1)
        raSource(i) = real(x1) 
        decSource(i) = real(y1)       

        !**************************************************
        ! Convert the standard coordinates of the catalogue
        ! objects to RA and DEC
        !**************************************************
        call sla_dtp2s(xye(1, i), xye(2, i), raPnt, decPnt, x, y)
        raCat(i) = real(x)
        decCat(i) = real(y)       
        distance(i) = ArcLength(dble(raSource(i)), dble(decSource(i)), dble(raCat(i)), dble(decCat(i)))
        rmsResidual = rmsResidual + distance(i) * distance(i)
    end do

    error = 1.0_double
    call ComputeAverage(distance, error, npts, mean, standardDeviation)
!    write(*, *) "Goodnessoffit- mean=", mean, " standard deviation=", standardDeviation

    rmsResidual = sqrt(rmsResidual / dble(npts))
    write(errstr, *) "RMS residual=", rmsResidual
    call message(AppMsg, Verbose)
              
    okay = .false.
    if (npts .gt. 20) then
        
        count = 0
        nSigma = 3
        rmsResidual = dble(nSigma) * rmsResidual
        do i = 1, npts
            if (distance(i) .gt. rmsResidual) count = count + 1
        end do
        write(errstr, *) "No of data rejected = ", count, "(", nSigma, " x rmsresidual)"
        call message(AppMsg, Verbose)
              
        !****************************************************
        ! Exclude data with residuals > nSigma * rmsResidual
        !****************************************************
        if (real(count) / real(npts) .lt. 0.25) then
            k = 0
            okay = .true.
            do i = 1, npts
                if (distance(i) .le. rmsResidual) then
                    k = k + 1
                    xye(1, k) = xye(1, i)
                    xye(2, k) = xye(2, i)
                    xym(1, k) = xym(1, i)
                    xym(2, k) = xym(2, i)
                end if 


            end do
            if (npts .eq. k) okay = .false.
            npts = k   ! The new number of points
        end if
    end if

    deallocate(raSource)
    deallocate(decSource)
    deallocate(raCat)
    deallocate(decCat)
    deallocate(distance)
    deallocate(error)
end subroutine GoodnessOfFit
!*****************************************************
! Subroutine AddCorrectionColumns
!*****************************************************
subroutine AddCorrectionColumns(fileName, size, raSource, decSource, &
                                raOffset, sigRaOffset,   &
                                decOffset, sigDecOffset, &
                                rotation, sigRotation,  maxLikelihood, noMatches)

    character(len=*), intent(in) :: fileName
    integer(kind=int32), intent(in) :: size
    real(kind=double), intent(in), dimension(1 : size) :: raSource
    real(kind=double), intent(in), dimension(1 : size) :: decSource 
    real(kind=single), intent(in) :: raOffset, sigRaOffset
    real(kind=single), intent(in) :: decOffset, sigdecOffset
    real(kind=single), intent(in) :: rotation, sigRotation
    real(kind=double), intent(in) :: maxLikelihood
    integer(kind=int32), intent(in) :: noMatches
    real(kind=single) :: nullValue

    type(DataSetT) :: set
    type(TableT)   :: sourceTable

    nullvalue = -real(99999)
    set = dataSet(fileName, MODIFY)
    sourceTable = table(set, "SRCLIST", MODIFY)

    call setAttribute(set, "RAOFFSET", raOffset, "arcsec", "RA correction applied")
    call setAttribute(set, "RAOFFERR", sigRaOffset, "arcsec", "Error in RA correction")
    call setAttribute(set, "DEOFFSET", decOffset, "arcsec", "Dec correction applied")
    call setAttribute(set, "DEOFFERR", sigDecOffset, "arcsec", "Error in dec correction")
    call setAttribute(set, "DEOFFERR", sigDecOffset, "arcsec", "Error in dec correction")
    call setAttribute(set, "ROT_CORR", rotation, "degrees", "Rotation correction applied")
    call setAttribute(set, "ROT_ERR", sigRotation, "degrees", "Error in rotation correction")
    call setAttribute(set, "NMATCHES", size, "number", "Number of sources matched")
    call setAttribute(set, "MAXLIKEL", maxLikelihood, "", "Maximum likelihood value")
       
    call AddReal64Column(sourceTable, raSource, size, "RA_CORR", "deg", nullValue)
    call AddReal64Column(sourceTable, decSource, size, "DEC_CORR", "deg", nullValue)

    call release(sourceTable)
    call release(set)

end subroutine AddCorrectionColumns

!****************************************************************
!
!
!****************************************************************
subroutine DrawGraph(ndata, xdata, xerrors, ydata, yerrors, &
         title, xlabel, ylabel, x1, x2, y1, y2)
  
    integer, intent(in) :: ndata
    real, intent(in) :: xdata(1 : ndata)
    real, intent(in) :: xerrors(1 : ndata)
    real, intent(in) :: ydata(1 : ndata)
    real, intent(in) :: yerrors(1 : ndata)

    character(len=*), intent(in) :: title
    character(len=*), intent(in) :: xlabel
    character(len=*), intent(in) :: ylabel
    character(len=3) :: string
    real, intent(in) :: x1, x2, y1, y2
    real, dimension( : ), allocatable :: lower, upper
    real, save ::  xmin, xmax, ymin, ymax
    integer i, j
       
    xmin =  10000.0
    xmax = -10000.0
    ymin = xmin
    ymax = xmax
             
    do j = 1, ndata
        
        if (xdata(j) .lt. xmin) xmin = xdata(j)
        if (xdata(j) .gt. xmax) xmax = xdata(j)
        if (ydata(j) .lt. ymin) ymin = ydata(j)
        if (ydata(j) .gt. ymax) ymax = ydata(j)
    end do
    if (abs(ymax - ymin) .lt. 1e-10) then
        ymin = ymin - 0.1
        ymax = ymax + 0.1
    end if    
    call pgsvp(x1, x2, y1, y2) 
    call pgrnge(xmin, xmax, xmin, xmax)
    call pgrnge(ymin, ymax, ymin, ymax)
    call pgwindow(xmin, xmax, ymin, ymax)
    call pgsch(1.0)
  
  
    call pgslw(4)
    call pgbox("bctsi", 0.0, 0, "bctsi", 0.0, 0)
    call pgsch(0.9)
    call pgslw(2)
    call pgbox("n", 0.0, 0, "n", 0.0, 0)

    call pgsch(1.2) 
    call pgslw(5)
    call pglabel(xlabel, ylabel, title)
    
    call pgslw(1)
  
    call pgsch(0.6)        
    call pgpoint(ndata, xdata, ydata, 9)
   
!    call pgline(ndata, xdata, ydata)

       
    allocate(lower(1 : ndata))
    allocate(upper(1 : ndata))
     
    lower = xdata - xerrors
    upper = xdata + xerrors
           
    call pgerrx(ndata, lower, upper, ydata, 0)
          
    lower = ydata - yerrors
    upper = ydata + yerrors
           
    call pgerry(ndata, xdata, lower, upper, 0)
    deallocate(lower)
    deallocate(upper)  

    call pgslw(5)
    call pgsch(0.5)
 
end subroutine DrawGraph
end module astromcorrect_mod
