!
!  Module      : omatt_mod
!
! Authors     : Tim Harries
!             : Andrew King (MSSL)
!          
!
! Description : Convert from OM detector to sky coordinates.
! 
! Comments    : 
!
! $Id: module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
! 
! History : Created  : March 98 - TJH
!           Modified : 21 May 98 - Added use sla.
!           Modified : 27 May 99 - Updated parameter names.
!           Modified : 1st June 00 - Added calculation of galactic coordinates for RA and DEC
!           Modified : 1st June 00 - Added addition of two extra keywords and columns to hold
!                                    the galactic coordinates.
!           Modified : 2nd June 00 - Changed the galactic coordinates to 8 byte data
!           Modified : 2nd June 00 - Installed code to release memory that had been allocated
!                                    this was left hanging allocated.
!           Modified : 4th june 00 - Changed RA and DEC to 8 byte data
!
!
!           Modified : 21/06/2000 Changed poserr to be consistent with arcsecs definition from OMDETECT.
!
!           Modified : 1/8/2000  (CB) - Added -90 degree image rotation
!                      to correct for error in coordinate system.
!           Modified : 14/8/2000 (CB) - Maths functions sin and cos replaced by dsin and 
!                                       dcos (double precision versions),
!                                       & some other changes to double precision made.
!           Modified : 15/8/2000 (CB) - Output image corrected for image distortion.
!           Modified : 21/8/2000 (CB) - Function "MissionTime" added to compute the difference
!                                       between the average of the start and end observations
!                                       currently fixed at 0.4765
!          Modified : 24/8/2000 (CB) -  Some write statements deleted. The +ve signs for 
!                                       the distortion corrections changed to -ve.
!          Modified : 07/09/2000     -  Attitude matrinningix read in. Boresight calculation changed -
!                                       uses [A*(B)^-1]*(1,0,0).
!          Modified : 28/09/2000(AMK)-  Changed parameters in call message!
!          modified : 02/10/2000(AMK)-  Reinstated the call to CAL_omGetPLateScalePnt
!
!          modified : 02/10/2000(SRR)-  Inserted inv_boremat for the inverted boresight matrix
!
!          modified : 05/10/2000(AMK)-  Replaced the call getFIlterID in mssllib by the OAL call filterIdToInt
!          modified : 10/10/2000(CB) -  Paramter "usecat" set to false if it is true- not yet implemented
!          modified : 20/12/2000(CB) -  Removed world-coordinate attributes from input image header file.
!          modified : 18/01/2000(CB) -  Exposure array added to the PPS output file. Code for astrometry
!                                       correction moved to a subroutine. 
!          modified : 24/05/2001(CB) -  Attributes "RA_SCX, DEC_SCX & PA added to output file- values
!                                       refer to the spacecraft pointing position. RA_PNT & RA_DEC
!                                       also added.  
!          modified : 14/06/2001(CB) -  Removed code that changes position errors (SPR 2429)
!          modified : 02/08/2001(CB) -  Removed definition of struct SpaceCraftAttitudeType-
!                                       defined in caloaldefs.  Replaced the matrix attmatrix in the 
!                                       call to OAL_toAttitudeMatrix by attmatrix1 (correct subscripts)
!                                       Changed program so that all 1-d arrays start at 1. Tidied
!                                       program up- removed astrometry correction code into a
!                                       separate module- added module "utility" which contains
!                                       some subroutines for doing routine jobs. 
! 
!          modified : 10.09.2001(VNY) - The centre of the CCD is fixed at (1024.5, 1024.5). Output RA and 
!                                       DEC coordinates has been recalculated according to the origin of the
!                                       CCD coordinate system lying at the center of the first pixel.
!                                       Correspondingly, the output image shift and rotation have
!                                       been corrected                            
!          modified : 16.11.2001 (CB) - Code to store an exposure-image map removed. Rotated image pixels
!                                       outside the image boundary set to NULL
!          modified : 16/12/2001(CB) -  Paramter "usecat" reactivated and astrometry correction now done
!                                       if sufficient stars   
!
!*****************************************************************************************  
module omatt_mod

  use param
  use cal
  use dal
  use sasdal
  use caloaldefs

  use errorHandling
  use sla
  use readModes_mod
  use allocation_mod 
  use utility_mod
  use sourcematch_mod
  use astromcorrect_mod    ! For doing the astrometry correction
implicit none
 
  private

   public omatt

contains

!---------------------------------------------------------------------
subroutine omatt
!---------------------------------------------------------------------

    type(DataSetT) :: inDataSet                                       ! Input dataset handle.
    type(DataSetT) :: sourceListDataSetH                              ! Source List dataset handle.
    type(DataSetT) :: PPSOSWDataSetH                                  ! PPS OSW dataset handle.
    type(ArrayT) :: inArrayH                                          ! Input array handle.
    type(ArrayT) :: PPSOSWArrayH                                      ! PPS OSW array handle.
    type(ArrayT) :: PPSOSWExpArrayH                                   ! Added by CB

    type(TableT) :: sourceListTableH                                  ! Source List table handle.

    type (SpacecraftAttitudeType) attitude

    character(len=255) :: set                                         ! OM image intermediate filename
    character(len=255) :: sourceListSet                               ! OM OSW Source List filename
    character(len=255) :: PPSOSWSet                                   ! PPS OM OSW FITS sky image filename.
    character(len=255) :: device                                      ! PGPLOT device.
    character(len=20):: rastring, decstring   ! strings for RA and DEC
    character(len=1) :: cs                    ! sign of dec +/-
    character(len=10)           :: datamode   ! datamode attribute
    character(len=6)            :: omFilter   ! the OM filter

    integer                    :: dim(2), nx, ny     ! size of the image
    integer                    :: xBinning
    integer                    :: yBinning
    integer             :: nxout, nyout        ! size of the output image
    integer             :: nrows              ! the number of rows in the table
    integer :: i, j, k                           ! counters
    integer             :: xbin, ybin         ! binning factors
    integer             :: xsize, ysize       ! size of the image
    integer             :: x0, y0             ! offset of the image
    integer             :: idmsf(4), ihmsf(4)  
    integer                     :: OM_MODE    ! the om mode

    real(kind=single), pointer :: oswImage(:, :)
    real                       :: refDX
    real                       :: refDY
    real                       :: rotatedRefDX
    real                       :: rotatedRefDY
    real(kind=double)  :: itrans(6)     ! its inverse

    ! sky coordinate image variables
    real(kind=single), pointer  :: outimage(:,:)       ! The output image
    real(kind=single)          :: exposureTime         ! The exposure time for an input image
    real(kind=double)  :: trans(6) ! rotation transformation matrix
                         
    real(kind=double)  :: rotationCorrection = 90.0_double * degreesToRadians ! Correction for coordinate error
    real(kind=double)  :: dXbinning, dYbinning
    real(kind=double)  :: sinRollAngle, cosRollAngle
   
    real(kind=single)  :: refx, refy         ! SRR added 4/10/00
    real(kind=double)  :: xmax,xmin,ymax,ymin            
    real(kind=single), dimension(:), pointer  ::  xpos
    real(kind=single), dimension(:), pointer  ::  ypos

    real(kind=double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(kind=double), dimension(:), pointer :: rightAscension
    real(kind=double), dimension(:), pointer :: declination
    real(kind=double), pointer :: RACorrected(:)
    real(kind=double), pointer :: decCorrected(:)


    real(kind=single), dimension(:), pointer :: posErr


    real(kind=single), dimension(:), pointer :: magnitude

    real(kind=double) :: ra_point, dec_point        ! position of the OM tangent point
    real(kind=double) :: rollangle                  ! the roll angle
    real(kind=double) :: time                        
    real(kind=double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(kind=double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(kind=double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(kind=double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(kind=double) :: pointvec(3)        ! s/c pointing
    real(kind=double) :: pointvec2(3)       
    real(kind=double) :: euler_axis(3)              ! euler axis for rotation
    real(kind=double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(kind=double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(kind=double) :: omvec(3)                   ! om pointing vector
    real(kind=double) :: rotmat(3, 3)               ! rotation matrix
    real, pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=double) :: xrad,yrad, rotxrad, rotyrad              ! position in radians     
    real(kind=double) :: r1                         ! modulus
    real(kind=double) :: currentEpoch
    real(kind=double) :: xposcorr, yposcorr
    real(kind=double), pointer :: xi_source(:)
    real(kind=double), pointer :: eta_source(:) 
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=double)           :: lambda           ! wavelength for CAL
    real(kind=double)           :: plateScale       ! the platescale
    real :: white, black
    real :: tr(6) = (/-0.5, 1.0, 0.0, -0.5, 0.0, 1.0/)
    real(kind=single) :: nullValue
    logical(kind=Bool) :: binBPE                    ! BPE binning on?
    logical(kind=bool) :: useCat                        ! Use the USNO-SA 1 catalog.
 
    integer                       :: istat              ! status for memory allocation
    integer(kind=int32)           :: filterIdInt

    real(kind=double), pointer :: galacticLong(:)   ! Array for galactic longitudes  
    real(kind=double), pointer :: galacticLat(:)    ! Array for galactic latitudes
   
    real(kind=single) ::  xShift, yShift, rotation, xScale, yScale     
    real(kind=single) ::  sigxShift, sigyShift, sigRotation     
    real(kind=double) ::  maxLikelihood     

    integer(kind=int32) :: noMatched
    logical :: status
  
    nullValue = real(-99999)
    !*************************************
    ! Get the command-line parameters
    !*************************************
    set = stringParameter("set")
    sourceListSet = stringParameter("sourcelistset")
    PPSOSWSet = stringParameter("ppsoswset")
    device = stringParameter("device")
    useCat = booleanParameter("usecat")


    !**************************
    ! Access the source list
    !**************************
    sourceListDataSetH = dataSet(sourceListSet, modify)      ! get osw list handle

   
    xbin = int32Attribute(sourceListDataSetH, "BINAX1")      ! binning factors
    ybin = int32Attribute(sourceListDataSetH, "BINAX2")
 
    if (xbin .ne. ybin) then
        call fatal("diffBin", "This version cannot deal with different binning factors in x and y")
    end if

    binBPE = booleanAttribute(sourceListDataSetH, "BINBPE")
 
    ! Calculate the image binning.
    xBinning = 2 ** xbin
    yBinning = 2 ** ybin

    if (binBPE) then
        xBinning = xBinning * 2
        yBinning = yBinning * 2
    end if

    dXbinning = dble(xBinning)  ! Need the double values of xBinning and 
    dYbinning = dble(yBinning)  ! yBinning fairly frequently

    omFilter = stringAttribute(sourceListDataSetH, "FILTER")
 

    dataMode = stringAttribute(sourceListDataSetH, "DATAMODE")  ! needed for cal

    if (trim(dataMode) .eq. "IMAGING") then 
       OM_MODE = IMAGE
    else
       OM_MODE = FAST
    end if
   
    sourceListTableH = table(sourceListDataSetH, "SRCLIST")     ! get the source list

    nrows = numberOfRows(sourceListTableH)

    !*********************************************************
    ! Get the x and y pixel coordinates of each source
    ! NB these are stored as real32 values.
    !********************************************************

    xpos => AllocateSingle(1, nRows, "xpos")
    call ReadReal32Column(sourceListTableH, "XPOS", nRows, 0, xpos)
    
 
    ypos => AllocateSingle(1, nRows, "ypos")
    call ReadReal32Column(sourceListTableH, "YPOS", nRows, 0, ypos)
 
    magnitude => AllocateSingle(1, nRows, "magnitude")
    poserr => AllocateSingle(1, nRows, "poserr")
 
    call ReadReal32Column(sourceListTableH, "MAG", nRows, 0, posErr)
    magnitude = real(posErr)      


    call ReadReal32Column(sourceListTableH, "POSERR", nRows, 0, posErr)
 


    ! the image contains the information on the science window

    inDataSet = DataSet(set, modify) 

    inArrayH = array(inDataSet, "PRIMARY", modify)

    exposureTime = real32Attribute(inDataSet, "EXPOSURE") ! Image exposure time


    !***********************************************************
    ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
    !***********************************************************
    call readModes(inDataSet, x0, y0, xsize, ysize)  

    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)

    !**********************************
    ! plot the image as a grayscale
    !**********************************
    call pgbegin(0, device, 1, 1)
    call pgenv(0.0, real(nx), 0.0, real(ny), 1, 0)

    white = minval(oswImage)
    black = maxval(oswImage)

    call pggray(oswImage, nx, ny, 1, nx, 1, ny, black, white, tr)

    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(sourceListDataSetH, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(sourceListDataSetH, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(sourceListDataSetH, "PA") * degreesToRadians       ! Spacecraft PA in radians
    rollAngle = pa_scx

   
    !************************************
    ! Get the filter Id integer value
    !************************************
    filterIdInt = filterIdToInt(omFilter)

    !*********************
    ! Set the CAL state
    !*********************
    call CAL_setState(OM, ccdModeId=OM_MODE, filterId = filterIdInt)

    call CAL_omGetPlateScale(lambda, plateScale)  ! Get the plate-scale (arcsecs per pixel) 
 
    call message(AppMsg, Verbose)

    plateScale = plateScale * degreesToRadians / 3600.0_double         ! platescale converted to radians


    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(sourceListDataSetH)

    currentEpoch = real32Attribute(sourceListDataSetH, "EQUINOX")

    !*********************************************************
    ! Need the current epoch for the astrometry correction
    !*********************************************************
    currentEpoch = 1998.0_double + time / (365.25_double * 86400.0_double)
    

    call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix
    
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)

    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle)

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)
    
    forall( i = 1 : 3) 
        forall(j = 1 : 3) attmatrix(i, j) = attmatrix1(i - 1, j - 1)
    end forall

    !**********************************************
    ! rotationCorrection corrects for OM axes definition.
    !**********************************************

    rollAngle = rollAngle + rotationCorrection
   
    xi_source => AllocateDouble(1, nRows, "xi_source")
    eta_source => AllocateDouble(1, nRows, "eta_source")
 
    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the mdulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !
    !
    ! use SIMON'S BIT
    !
    !write(*,*)'using SIMONS BIT (essentially redundant now but keep for now
    !--------------------------------------------------------------------------------------------------------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !--------------------------------------------------------------------------------------------------------

    euler_axis = -rollangle * pointvec        !leave in (SRR)

    ! convert euler axis vector to a rotation matrix
    call sla_dav2m(euler_axis, rotmat) 

    ! initialise output vector
    omvec = 0.0_double


!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************

    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)

    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************

    call sla_dcc2s(omvec, ra_point, dec_point)

    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi


    write(errstr, '(a, f8.4)') "RA of OM boresight: ", ra_point * radiansToDegrees
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", dec_point * radiansToDegrees
    call message(AppMsg, Verbose)


    deltax => AllocateSingle(0, 0, "deltax")
    deltay => AllocateSingle(0, 0, "deltay")
 

    !*********************************************************************
    ! Loop over source list and compute the undistorted position of
    ! each star realtive to the centre of the full-frame image, and also
    ! the standard coordinates.
    !*********************************************************************

    sinRollAngle = dsin(rollAngle)  ! Need these quantities quite a lot
    cosRollAngle = dcos(rollAngle)
   
    rightAscension => AllocateDouble(1, nRows, "rightAscension")
    declination => AllocateDouble(1, nRows, "declination")
 

    do i = 1, nrows

        !****************************************************
        ! compute absolute position of object within OM frame
        !****************************************************
        xposcorr = dble((xpos(i) - 1.0)) * dXbinning + dble(x0)
        yposcorr = dble((ypos(i) - 1.0)) * dYbinning + dble(y0)

        write(errstr, '(a,i4.4,",",i4.4,a)') &
            "Centroided pixel position (", nint(xposcorr), nint(yposcorr),")"
        call message(AppMsg, Verbose)

        !***************************************************
        ! correct positions for pincushion/barrel distortion
        !***************************************************

        rawx(0) = real(xposcorr)
        rawy(0) = real(yposcorr)

        call CAL_omDistortion(rawx, rawy, deltax, deltay)

        xposcorr = xposcorr - dble(deltax(0))    
        yposcorr = yposcorr - dble(deltay(0))
      
        !***************************************************
        ! get offset relative to tangent point (at OM centre)
        !***************************************************
        xrad = xposcorr - xCentre 
        yrad = yposcorr - yCentre

        !**************************************************
        !  rotate x,y frame to tangent sky (RA, DEC) frame
        !*************************************************
        rotxrad = plateScale * (xrad * cosRollAngle - yrad * sinRollAngle)
        rotyrad = plateScale * (xrad * sinRollAngle + yrad * cosRollAngle)

        !****************************************************
        ! store standard coords for use astrometry correction
        !****************************************************
        xi_source(i)  = -rotxrad   ! Tangent-plane coords
        eta_source(i) =  rotyrad 
   
    end do

    
    !***********************************************************
    ! Transform tangent-plane coordinates (xi_source, yi_source) 
    ! into spherical (sky) coordinates (RA, dec). ra_point,
    ! dec_point is the spherical coordinate of the tangent-point
    ! Also calculate the galactic longitude and latitude of
    ! each source
    !************************************************************
   
    galacticLong => AllocateDouble(1, nrows, "galacticLong")
    galacticLat => AllocateDouble(1, nrows, "galacticLat")
 
 
    do i = 1, nrows
        call sla_dtp2s(xi_source(i), eta_source(i), &
            ra_point, dec_point, rightAscension(i), declination(i))

        call sla_EQGAL(rightAscension(i),  declination(i), &
                      galacticLong(i), galacticLat(i))
    end do

    !********************************
    ! convert to degrees
    !********************************
    RACorrected => AllocateDouble(1, nrows, "RA Corrected")
    decCorrected => AllocateDouble(1, nrows, "Dec Corrected")
    RACorrected = rightAscension
    decCorrected = declination


    rightAscension = rightAscension * radiansToDegrees
    declination = declination * radiansTodegrees
    galacticLong = galacticLong * radiansToDegrees
    galacticLat  = galacticLat * radiansToDegrees

    !*********************************************************
    ! Add RA, dec, galactic longitude and galactic latitude 
    ! columns to the fits source-list file
    !*********************************************************
    call AddReal64Column(sourceListTableH, rightAscension, nrows, "RA", "Deg", nullvalue)
    call AddReal64Column(sourceListTableH, declination, nrows, "DEC", "Deg", nullValue)
    call AddReal64Column(sourceListTableH, galacticLong, nrows, "LII", "Deg", nullValue)
    call AddReal64Column(sourceListTableH, galacticLat, nrows, "BII", "Deg", nullValue)


    !**********************************************************
    ! We only do the next bit if we have access to the usno ccd
    !**********************************************************

    if (useCat) then

        call CorrectAstrometry(nrows, RACorrected, decCorrected, &
            posErr, magnitude, ra_point, dec_point,  sourceListSet, &
            status, xShift, sigxShift, yShift, sigyShift, &
            rotation, sigRotation, xScale, yScale, maxLikelihood, noMatched)     
    
        if (status) then  ! The astrometry correction was successful
            raCorrected  = raCorrected * radiansToDegrees
            decCorrected = decCorrected * radiansToDegrees

            call AddCorrectionColumns(sourceListSet, nrows, raCorrected, decCorrected, &
                                xShift, sigxShift, yShift, sigyShift, rotation, &
                                sigRotation, maxLikelihood, noMatched)
        end if
    end if




    deallocate(rightAscension, stat = istat)
    call CheckDeallocation(istat, "rightAscension")

    deallocate(declination, stat = istat)
    call CheckDeallocation(istat, "declination")

    deallocate(posErr, stat = istat)
    call CheckDeallocation(istat, "posErr")

    deallocate(magnitude, stat = istat)
    call CheckDeallocation(istat, "magnitude")

    deallocate(galacticLong, stat = istat)
    call CheckDeallocation(istat, "galacticLong")

    deallocate(galacticLat, stat = istat)
    call CheckDeallocation(istat, "galacticLat")
 
    call SetFitsTableAttributes(sourceListDataSetH, xBinning, yBinning, plateScale, refx, refy, &
                                  ra_scx, dec_scx, ra_point, dec_point)


    call release(sourceListTableH)
    call release(sourceListDataSetH)

    !****************************************************************
    !  NOW CREATE THE OUTPUT PPS PRODUCT OM OSW SKY FITS IMAGE. 
    !  This is done by resampling the detector image on a sky grid. 
    ! the transformation is currently a simple rotation
    ! Define the transformation terms (note the offsets will change)
    !****************************************************************

    trans(1) =  0.0_double
    trans(2) =  dcos(-rollangle)
    trans(3) =  dsin(-rollangle)
    trans(4) =  0.0_double
    trans(5) = -dsin(-rollangle)
    trans(6) =  dcos(-rollangle)

    !*******************************************
    ! create the output source file
    !*******************************************
    PPSOSWDataSetH = DataSet(PPSOSWSet, create)

    !**********************************************************************
    ! Determine the bounding x and y (xmin, xmax, ymin, ymax) coordinates 
    ! of the image
    !*********************************************************************
    call  GetImageBoundaries(nx, ny, real(x0), real(y0), xBinning, yBinning, &
                             trans, xMin, xMax, yMin, yMax)


! invert transformation matrix for later use

    call invert_trans(trans, itrans)    ! need inverted transformation matrix later to rotate image

    nxout = xmax - xmin                  ! determine image dimensions
    nyout = ymax - ymin

!***********************************************************************************
! set up the output image with the appropriate size (xmax,ymax already referenced to 
! origin of containing (output) box so these equal the output box dimensions).
!**********************************************************************************

    dim(1) = nxout     ! The dimensions of the output image
    dim(2) = nyout
    PPSOSWArrayH    = addArray(PPSOSWDataSetH, "PRIMARY",  Real32, dim)  ! For the image array


    ! copy the header across

    call copyAllAttributes(attributable(inDataSet), attributable(PPSOSWDataSetH))
   

    ! add content keyword

    call setAttribute(PPSOSWDataSetH, "CONTENT", "OM OSW FITS SKY IMAGE", &
         "Content of FITS file")

!********************************************
! convert from ra and dec numbers to strings
!********************************************

    call SLA_CR2TF(2, real(ra_point), cs, ihmsf)
    write(rastring, '(i2.2,'':'',i2.2,'':'',i2.2,''.'',i2.2)') &
         ihmsf(1), ihmsf(2), ihmsf(3), ihmsf(4)
    call SLA_CR2AF(1, real(dec_point), cs, idmsf)
    write(decstring, '(a1,i2.2,'':'',i2.2,'':'',i2.2,''.'',i1)') &
         cs, idmsf(1), idmsf(2), idmsf(3), idmsf(4)

!******************************************************************************
! now determine refDX, refDY offsets of OM plate centre (PC) from original OSW 
! origin (IN BINNED PIXELS)
!******************************************************************************


    refDX = (xCentre - x0) / dXbinning
    refDY = (yCentre - y0) / dYbinning


! Compute these offsets in terms of the rotated frame (remember rotation 
! takes place about the OM plate centre so the x,y offsets of the OSW origin 
! shift in the rotated frame). 


    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

! Finally compute offsets (refx, refy) of the OM plate centre from the BB origin, i.e. 
! just add the offset of the PC from the OSW origin to the offset of the OSW origin 
! from the BB origin

    refx = trans(1) + rotatedRefDX
    refy = trans(4) + rotatedRefDY
!
! And write the WCS attributes to output file. Note that the scale is in binned pixels (ie 
! if binning is 2, with raw scale -0.5" per pixels, CDELT1/2 =1
!


    !****************************************************
    ! Set the fits header attributes for the image array
    !****************************************************
    call SetFitsArrayAttributes(PPSOSWArrayH, xBinning, yBinning, plateScale, refx, refy, &
                                  ra_scx, dec_scx, ra_point, dec_point)

    outimage => real32array2data(PPSOSWArrayH)

 
    !*********************************************************  
    !  Now rotate the image.  
    !*********************************************************

    k = 0
    do i = 0, nxout - 1
        do j = 0, nyout - 1
            call setToNull( PPSOSWArrayH, k)

            k = k + 1
        end do
    end do


    call RotateImage(nx, ny, dXbinning, dYbinning, x0, y0, nxout, nyout, &
                     oswImage, outImage, itrans)  

 

    ! Release handles

    call release(inArrayH)
    
    call release(PPSOSWArrayH) 
    
    call release(PPSOSWDataSetH)
    call release(inDataSet)

    !**************************************
    ! Release memory allocation
    !**************************************

    deallocate(xpos, stat = istat)
    call CheckDeallocation(istat, "xpos")

    deallocate(ypos, stat = istat)
    call CheckDeallocation(istat, "ypos")

    deallocate(xi_source, stat = istat)
    call CheckDeallocation(istat, "xi-source")

    deallocate(eta_source, stat = istat)
    call CheckDeallocation(istat, "eta_source")

    deallocate(deltax, stat = istat)
    call CheckDeallocation(istat, "deltax")
 
    deallocate(deltay, stat = istat)
    call CheckDeallocation(istat, "deltay")
 
end subroutine omatt
!******************************************
! invert_trans
! inverts the transformation matrix, i.e. gets coefficients to swap between
! x'= t(1) + t(2)*x + t(3)*y
! y'= t(4) + t(5)*x + t(6)*y
! and 
! x= it(1) + it(2)*x' + it(3)*y'
! y= it(4) + it(5)*x' + it(6)*y'
!******************************************
subroutine invert_trans(trans, itrans)
    real(kind=double), intent(in) ::  trans(6)
    real(kind=double), intent(out) :: itrans(6)
    real(kind=double) determ

    determ = trans(2) * trans(6) - trans(3) * trans(5)
    itrans(1) = trans(3) * trans(4) - trans(1) * trans(6)
    itrans(2) = trans(6)
    itrans(3) = -trans(3)
    itrans(4) = trans(1) * trans(5) - trans(2) * trans(4)
    itrans(5) = -trans(5)
    itrans(6) = trans(2)
    itrans = itrans / determ

end subroutine invert_trans
!*******************************************
! Function MissionTime - Returns the number
! of seconds from the start of the mission
! to the observation time. Added by CB because
! the call to CAL_GetBoresiteMatrix needs
! the time in seconds from the launch.
!******************************************
double precision function  MissionTime(primaryTable)
    type(DataSetT), intent(in)   :: primaryTable

    character(len = 19)  :: dateObs         
    character(len = 19)  :: dateEnd        
    real(kind = double)  :: JulianDate0, JulianDate1, JulianDate2
    character(len = 19)  :: missionReferenceTime = "1998-01-01T00:00:00"

    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the start of the observations.
    dateObs = stringAttribute(primaryTable,  "DATE-OBS")    
    JulianDate1 = ModifiedJulianDate(dateObs)

    ! Compute the Julian date of the end of the observations.
    dateEnd = stringAttribute(primaryTable,  "DATE-END")   
    JulianDate2 = ModifiedJulianDate(dateEnd)

    ! The required time is the difference between the average of the 
    ! start and end times and the mission reference time, in seconds.
    MissionTime = ((JulianDate1 + JulianDate2) / 2.0d0 - JulianDate0) * 86400.0d0

end function MissionTime
!*******************************************************************************
! subroutine GetImageBoundaries
! The first step is to find the limits of the sky box that will bound the rotated 
! image, after it is linearised (distortion corrections applied). This is done by 
!
! 1) computing the raw coordinates of the image corners and applying the appropriate 
! distortion correction to them. The coordinates are referenced back to the origin 
! of the undistorted image. The corners are then rotated by the roll angle and 
! converted to binned pixels 
! 2) the limits of this rotated, distorted image are obtained and the translational 
! shift betwen the bounding box origin and the OSW origin are passed to the 
! transformation matrix leading coefficients.
! 3) The dimensions of the bounding box are determined
! 4) The offset of the OM plate centre (1024,1024) are determined relative to 
! the OSW original, as meadured in the rotated frame, and then this is 
! corrected to be wrt to the bounding box origin
! 5) WCS keywords are set
! 6) pixels in the sky box are now converted, via the inverse transformation matrix, 
! to the OSW frame and an iterative process to correct back from linearised to raw 
! pixels is used to find the original pixel in the raw image. Bi-linear interpolation 
! is used to fill the rotated image  
!*******************************************************************************
subroutine GetImageBoundaries(nx, ny, x0, y0, xBinning, yBinning, &
                          trans, xMin, xMax, yMin, yMax)
    integer, intent(in) :: nx, ny   ! The image dimensions
    real(kind=single),  intent(in) :: x0, y0
    integer, intent(in) :: xBinning, yBinning
    real(kind=double), intent(inout), dimension(1 : 6) :: trans   
    real(kind=double), intent(out) :: xMin, xMax, yMin, yMax
    real(kind=single) :: rawx(0 : 0), rawy(0 : 0)      ! CAL coords
    real(kind=single), pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=double)  :: xlin(4), ylin(4), xcorn(4), ycorn(4)

! (30/08/01) NOTE definition of CAL_omDistortion means we could set up rawx/y arrays 
! first and then call distortion routine once (passing all corners in at once) and then 
! put loop after to compute lin coords (SRR).
     
! bottom left hand corner

    rawx(0) = x0                              !at 2048 pixels resolution 
    rawy(0) = y0
    call CAL_omDistortion(rawx, rawy, deltax, deltay)   
    xlin(1) = rawx(0) - deltax(0) - x0            !correct for distortion and reference to x0,y0 origin
    ylin(1) = rawy(0) - deltay(0) - y0
    xcorn(1) = (trans(1) + trans(2) * xlin(1) + trans(3) * ylin(1)) / dble(xBinning)      !rotate + convert to binned pixels
    ycorn(1) = (trans(4) + trans(5) * xlin(1) + trans(6) * ylin(1)) / dble(yBinning)

!    write(*,*)'1 rawx,rawy = ',rawx(0),rawy(0)
!    write(*,*)'1 dx,dy = ',deltax(0),deltay(0)
!    write(*,*)'1 xl,yl= ',xlin(1),ylin(1)
!    write(*,*)'1 xc,yc = ',xcorn(1),ycorn(1)

!bottom rh corner in 2048 pixels

    rawx(0) = x0 + real(nx * xBinning)      
    rawy(0) = y0
    call CAL_omDistortion(rawx, rawy, deltax, deltay)   
    xlin(2) = rawx(0) - deltax(0) - x0
    ylin(2) = rawy(0) - deltay(0) - y0
    xcorn(2) = (trans(1) + trans(2) * xlin(2) + trans(3) * ylin(2)) / dble(xBinning)
    ycorn(2) = (trans(4) + trans(5) * xlin(2) + trans(6) * ylin(2)) / dble(yBinning)

! top rh corner

    rawx(0) = x0 + real(nx * xBinning)
    rawy(0) = y0 + real(ny * yBinning)
    call CAL_omDistortion(rawx, rawy, deltax, deltay)   
    xlin(3) = rawx(0) - deltax(0) - x0               
    ylin(3) = rawy(0) - deltay(0) - y0
    xcorn(3) = (trans(1) + trans(2) * xlin(3) + trans(3) * ylin(3)) / dble(xBinning)
    ycorn(3) = (trans(4) + trans(5) * xlin(3) + trans(6) * ylin(3)) / dble(yBinning)

! top lh corner

    rawx(0) = x0
    rawy(0) = y0 + real(ny * yBinning)
    call CAL_omDistortion(rawx, rawy, deltax, deltay)   
    xlin(4) = rawx(0) - deltax(0) - x0
    ylin(4) = rawy(0) - deltay(0) - y0
    xcorn(4) = (trans(1) + trans(2) * xlin(4) + trans(3) * ylin(4)) / dble(xBinning)
    ycorn(4) = (trans(4) + trans(5) * xlin(4) + trans(6) * ylin(4)) / dble(yBinning)


!  We now have the corners of OSW in linearised coordinates relative to the origin 
!  of the original OSW image, after rotation of the frame 
!  Now need to get limits of bounding box (BB)  from these corners (in rotated frame)
!  Once minimum x and y are found (origin of BB - i.e. the rotated frame), since these 
!  are wrt the original OSW origin, we can determine the offset of the OSW origin from 
!  bounding BB (transfer offsets to trans(1), trans(4))

    xmax = trans(1)
    xmin = trans(1)
    call GetMinMax(xmin, xmax, xcorn)
    trans(1) = trans(1) - xmin                        !get x offset of OSW origin from BB origin

    ymax = trans(4)
    ymin = trans(4)
    call GetMinMax(ymin, ymax, ycorn)
    
    trans(4) = trans(4) - ymin                        !get y offset of OSW origin from BB origin

end subroutine GetImageBoundaries
!******************************
! subrioutine GetMinMax
!******************************
subroutine GetMinMax(minimum, maximum, corner)
    real(kind=double), intent(inout) :: minimum
    real(kind=double), intent(inout) :: maximum
    real(kind=double), intent(in), dimension(1 : 4) :: corner
    integer :: i

    do i = 1, 4
        minimum = dmin1(minimum, corner(i))
        maximum = dmax1(maximum, corner(i))
    end do

end subroutine GetMinMax 
!*****************************************
! subroutine SetFitsArrayAttributes
!*****************************************
subroutine SetFitsArrayAttributes(arrayH, xBinning, yBinning,     &
                                  plateScale, refx, refy, ra_scx, &
                                  dec_scx, ra_point, dec_point)

    type(ArrayT), intent(in) :: arrayH   ! Array handle.
    integer, intent(in) :: xBinning, yBinning 
    real(kind=double), intent(in) :: plateScale
    real(kind=single), intent(in) :: refx, refy
    real(kind=double), intent(in) :: ra_scx, dec_scx
    real(kind=double), intent(in) :: ra_point, dec_point


    call setAttribute(arrayH, "CTYPE1", "RA---TAN", "Coordinate system for X")
    call setAttribute(arrayH, "CRPIX1", refx, "WCS reference pixel") 
    call setAttribute(arrayH, "RA_SCX", ra_scx * radiansToDegrees, &
                              "Coordinate of spacecraft pointing direction for X")
    call setAttribute(arrayH, "CRVAL1", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "CDELT1", (-1.0)*plateScale*(real(xBinning) * radiansToDegrees), &
                              "WCS pixel size")
    call setAttribute(arrayH, "CTYPE2", "DEC--TAN", "Coordinate system for Y")
    call setAttribute(arrayH, "CRPIX2", refy, "WCS reference pixel")
    call setAttribute(arrayH, "DEC_SCX", dec_scx * radiansToDegrees, &
                              "Coordinate of spacecraft pointing direction for Y")
    call setAttribute(arrayH, "CRVAL2", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
    call setAttribute(arrayH, "CDELT2", plateScale*(real(yBinning)*radiansToDegrees), &
                              "WCS pixel size")
    call setAttribute(arrayH, "RA_PNT", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "DEC_PNT", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
 
end subroutine SetFitsArrayAttributes


!*****************************************
! subroutine SetFitsTableAttributes
!*****************************************
subroutine SetFitsTableAttributes(arrayH, xBinning, yBinning,     &
                                  plateScale, refx, refy, ra_scx, &
                                  dec_scx, ra_point, dec_point)

    type(DataSetT), intent(in) :: arrayH   ! Array handle.
    integer, intent(in) :: xBinning, yBinning 
    real(kind=double), intent(in) :: plateScale
    real(kind=single), intent(in) :: refx, refy
    real(kind=double), intent(in) :: ra_scx, dec_scx
    real(kind=double), intent(in) :: ra_point, dec_point


!    call setAttribute(arrayH, "CTYPE1", "RA---TAN", "Coordinate system for X")
!    call setAttribute(arrayH, "CRPIX1", refx, "WCS reference pixel") 
!    call setAttribute(arrayH, "RA_SCX", ra_scx / DegtoRad, &
!                              "Coordinate of spacecraft pointing direction for X")
!    call setAttribute(arrayH, "CRVAL1", ra_point / DegtoRad, &
!                              "Coordinate of pointing direction for X")
!    call setAttribute(arrayH, "CDELT1", (-1.0)*plateScale*(real(xBinning)/degtorad), &
!                              "WCS pixel size")
!    call setAttribute(arrayH, "CTYPE2", "DEC--TAN", "Coordinate system for Y")
!    call setAttribute(arrayH, "CRPIX2", refy, "WCS reference pixel")
!    call setAttribute(arrayH, "DEC_SCX", dec_scx / DegtoRad, &
!                              "Coordinate of spacecraft pointing direction for Y")
!    call setAttribute(arrayH, "CRVAL2", dec_point / DegtoRad, &
!                              "Coordinate of pointing direction for Y")
!    call setAttribute(arrayH, "CDELT2", plateScale*(real(yBinning)/degtorad), &
!                              "WCS pixel size")
    call setAttribute(arrayH, "RA_PNT", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "DEC_PNT", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
 
end subroutine SetFitsTableAttributes


!***************************************************************
! Subroutine RotateImage
! Thius is done by-
! 1) looping over regular grid of output (sky) image points, 
! 2) computing (linear) location in unrotated image (via inverse transform)
! 3) correcting linear position in unrotated frame for distortion (i.e.  
!     converting back to raw pixel position) (2 stage iteration)
! 4) bi-linearly interpolating surrounding points 
!
!****************************************************************
subroutine  RotateImage(nx, ny, dXbinning, dYbinning, x0, y0, & 
                      nxout, nyout, oswImage, rotatedImage, &
                      itrans)  
    integer, intent(in) :: nx, ny
    real(kind=double), intent(in) :: dXbinning, dYbinning
    integer, intent(in)             :: x0, y0             ! offset of image
    integer, intent(in) :: nxout, nyout
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: oswImage
    real(kind=single), intent(out), dimension(0 : nxout - 1, 0 : nyout - 1) :: rotatedImage
    real(kind=double), intent(in)  :: itrans(6)     ! its inverse
  
    !*******************************
    ! Local variables
    !*******************************
    real(kind=double)  :: t, u, rx0, ry0, rx, ry
    real               :: rxInit, ryInit
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=double), dimension(:), pointer :: itrans3TimesjPlusPoint5  ! Stores itrans(3) * (j + 0.5)
    real(kind=double), dimension(:), pointer :: itrans6TimesjPlusPoint5  ! Stores itrans(6) * (j + 0.5)
    real, pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=double) :: dblNx, dblNy
    real(kind=double) :: radius1, radius2, oldDiff, newDiff
    real(kind=double) :: rxOld, ryOld
    integer :: i, j, j0, k0 
    integer :: j1, k1, iter, istat
    integer :: nxMinus2, nyMinus2
    integer :: maxIterations 


    nxMinus2 = nx - 2
    nyMinus2 = ny - 2
    maxIterations = 3

    itrans3TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans3TimesjPlusPoint5")
    itrans6TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans6TimesjPlusPoint5")
    
    !***************************************************************
    ! The following is to reduce execution time when creating the
    ! rotated image
    !***************************************************************
    forall (j = 0 : nyout - 1) 
        itrans3TimesjPlusPoint5(j) = itrans(3) * (dble(j) + 0.5_double)
        itrans6TimesjPlusPoint5(j) = itrans(6) * (dble(j) + 0.5_double)
    end forall


    dblNx = dble(nx)
    dblNy = dble(ny)

    do i = 0, nxout - 1
       rx0 = itrans(1) + itrans(2) * (dble(i) + 0.5_double)
       ry0 = itrans(4) + itrans(5) * (dble(i) + 0.5_double)
     
       do j = 0, nyout - 1

          !***************************************************************      
          !  use inverse transformation to get coords in unrotated 
          ! (input) frame note part of this is already done within i loop
          !***************************************************************
          rx = rx0 + itrans3TimesjPlusPoint5(j)
          ry = ry0 + itrans6TimesjPlusPoint5(j)

!          if(i.eq.17.and.j.eq.1019)write(*,*)' after first rx,ry calc rx,ry = ',rx,ry

          !***************************************************************************
          !  Calculate position in full frame. Get distortion correction and correct 
          !  positions. THIS is done iteratively (twice) because we want to go from 
          !  linearised to raw position but CAL distortion correction routine gives 
          !  values based on raw input positions. Thus we correct linearised position 
          !  with wrong correction values. To improve the final position we apply 
          !  the wrong distortion correction to get new coordinates and then recompute 
          !  the corrections using this position. This should bring us close to the 
          !  true raw position. 
          !
          ! VNY 10.09.2001 correction for the half of the pixel as the CCD 
          ! grid passes through the pixel center
          !****************************************************************** 
        
          rxInit = rx
          ryInit = ry
          oldDiff = 1e6
          do iter = 1, maxIterations
             rawx(0) = (rx - 0.5) * dXbinning + x0       ! position in full OM field.
             rawy(0) = (ry - 0.5) * dYbinning + y0

!             if(i.eq.17.and.j.eq.1019)write(*,*)' iter = ',iter,' rawx,rawy= ',rawx(0),rawy(0)
 
             call CAL_omDistortion(rawx, rawy, deltax, deltay)

!             if(i.eq.17.and.j.eq.1019)write(*,*)' after dist corr dx,dy= ',deltax,deltay

             rx = rxInit + deltax(0) / dXbinning   ! 25.8.2000 The signs are +ve because we are going from linear 
             ry = ryInit + deltay(0) / dYbinning   ! to raw positions. Convert back to binned pixels.

!             if(i.eq.17.and.j.eq.1019)write(*,*)' corrected posn rx,ry= ',rx,ry

             radius1 = sqrt(rx * rx + ry * ry)
             radius2 = sqrt(rxInit * rxInit + ryInit * ryInit)
             newDiff = abs(radius1 - radius2)

!             if(i.eq.17.and.j.eq.1019)write(*,*)' rad1,2,newdiff,olddiff= ',radius1,radius2,newdiff,olddiff

             if (newDiff .gt. oldDiff) then
                 rx = rxOld
                 ry = ryOld

                 exit
             else  
                 if (newDiff  .lt. 0.05) then
                 
                     exit
                 
                 endif 
             end if
             oldDiff = newDiff
             rxOld = rx
             ryOld = ry
          end do

          !*********************************************
          ! use bi-linear interpolation to fill in grid
          !*********************************************
          if (rx .ge. 0.0_double .and. rx .le. dblNx .and. & 
              ry .ge. 0.0_double .and. ry .le. dblNy) then

              !********************************************************************
              !   VNY 10.09.2001 correction for the half of the pixel as the CCD 
              !   grid passes through the pixel center
              !********************************************************************
              rx = rx - 0.5_double
              ry = ry - 0.5_double


!              if(i.eq.17.and.j.eq.1019)write(*,*)' rx,ry = ',rx,ry,' nxminus2,nyminus2=',nxminus2,nyminus2

              j0 = max(0, int(rx))
              j0 = min(nxMinus2, j0)
              k0 = min(nyMinus2, int(ry))
              k0 = max(0, k0)
              j1 = j0 + 1
              k1 = k0 + 1

              t = (rx - dble(j0))  
              u = (ry - dble(k0))

              rotatedImage(i, j ) = (1.0_double - t) * (1.0_double - u) * oswImage(j0, k0) + &
                                    t                * (1.0_double - u) * oswImage(j1, k0) + &
                                    t                * u                * oswImage(j1, k1) + &
                                   (1.0_double - t)  * u                * oswImage(j0, k1)


          end if
       end do
    end do

    deallocate(itrans3TimesjPlusPoint5, stat = istat)
    call  CheckDeAllocation(istat, "itrans3TimesjPlusPoint5")
 
    deallocate(itrans6TimesjPlusPoint5, stat = istat)
    call  CheckDEAllocation(istat, "itrans6TimesjPlusPoint5")
 

end subroutine RotateImage
end module omatt_mod





























































































































































































































































